/* QLAN to USB Adapter v2.1f, 20210107 for Teensy (++2.0)
 * Detect and capture network traffic on the 1-wire QLAN port, forwarding to host PC via USB-SER interface
 * - 1.0: Message Management/RESET & STATUS
 * - 1.1: Basic SENDing capability
 * - 1.2: Basic RECeiVing capability
 * - 1.8: Extended SCOUT to 585us (from 550) when sending to better suit ZXNet/Spectrum which doesn't seem to catch our transmissions...
 *      : Reverted SCOUT length - turns-out it was the bit-length of the ACK sent back.
 * - 1z9a:  ZX and QL versions (separate for now) with slightly different SEND bit-timings.
 *          - RECV bit-tmings has been normalised to ZX (13/21) which still works for QL (11/18)
 *
 * - 1z9b: Added DEBUG_STROBE_TIME to ensure that DEBUG_NET1 stays active long enough to register (>1us)
 * - 1z9c: Slightly increased RECV_BIT_TIME (13 to 14) to better capture ZX Net byte-frame
 * - 1z9d: Slight adjustment to DEBUG_NET2 and NET3 outputs (they were previously left high between packet-segments)
 * - 1x9d: Attempts to match default QXL transmission timings (10.67 us bit-time)
 *
 * - 2.0a: Reverted the QXL timing changes introduced in 1x9d - timing since corrected in QXL itself
 *         Added MQ_ERROR_MASK as bit 15 in M_ID (M_ID can now fall between 1-8191)
 *         - Added new DEBUG_NETx diagnostic outputs:
 *            NET_IN     = NetIN                                = PORTD-0 Grey
 *            NET_OUT    = NetOUT                               = PORTD-7 Brown
 *            DEBUG_NET1 = NetIN sample Strobe                  = PORTD-1 Purple
 *            DEBUG_NET2 = NetIN SCOUT/start/stop-bit Sample    = PORTD-2 Blue
 *            DEBUG_NET3 = NetIN data-bit Sample                = PORTD-3 Green
 *
 *            DEBUG_NET4 = NET RECV Error Strobe                = PORTD-4 Yellow
 *                           5us: RECV_NO_GAP
 *                          15us: RECV_NO_SCOUT
 *                          25us: RECV_HDR_FRAME_ERROR
 *                          35us: RECV_HDR_CHKSUM_ERROR
 *                          45us: RECV_HDR_NOT_FOR_ME
 *                          55us: RECV_HDR_PHANTOM_BLOCK
 *                          65us: RECV_HDR_BLK_MISMATCH
 *                          75us: RECV_DATA_FRAME_ERROR
 *                          85us: RECV_DATA_CHKSUM_ERROR
 *                          95us: RECV_DATA_ACK_NO_GAP
 *                         105us: RECV_OTHER_ERROR
 *
 *            DEBUG_NET5 = Net SEND Error Strobe                = PORTD-5 Orange
 *                           5us: SEND_NO_GAP
 *                          15us: SEND_CONTENTION 
 *                          25us: SEND_NO_HACK
 *                          35us: SEND_NO_DACK
 *
 *            DEBUG_NET6 = (currently unused)                   = PORTB-7 White
 *
 * - 2.1a: More versatile Buffer/Heap handling
 *
 * - 2.1b - Added 'Listen to self', 'Broadcast RECV' and FSERVE netHDR matching, as well as FSERVE blkL/blkH handling correction
 *          Adjusted for FSERVE-style Data ChkSums
 *          Both FSERVE packet receive, as well as Broadcast Send/Recv now complete :-)
 *
 * - 2.1e - Corrected Phantom Block handling - packet is now read, ACK'd and only then discarded
 *
 * - 2.1f - Reverted bit-timing constants to suit default QL/TK2 setup
 *
 * -TODO:
 *
 *          Transition to hardware timers (timer2 - only conflicts with __tone__ which we don't use)
 *          This is to replace the use of timed-delays. See https://arduino-info.wikispaces.com/Timers-Arduino
 *          TCCRx - Timer/Counter Control Register. The prescaler can be configured here.
 *          TCNTx - Timer/Counter Register. The actual timer value is stored here.
 *          OCRxy - Output Compare Register
 *          ICRxy - Input Capture Register (only for 16bit timer)
 *          TIMSKx - Timer/Counter Interrupt Mask Register. To enable/disable timer interrupts.
 *          TIFRx - Timer/Counter Interrupt Flag Register. Indicates a pending timer interrupt.

*/

elapsedMicros timer, ledTimer;
// elapsedMillis ledTimer;

// Version constant
const char QLAN_USB_VERSION[]="2.1f - 2021-01-07"; // NB - Remember to change this with each new version!

//Heap Handling constants
const signed int HEAP_SIZE=4096+2;
const signed int HEAP_HDR_SZ=4;
const signed int HEAP_ERR_IMEM=-5;
const signed int HEAP_ERR_ITNF=-7;
const signed int HEAP_END=0;
const signed int HEAP_FREE_SPC_PTR=0;

// Some timer constants

// const byte debugTimingEntries=31;

const unsigned long MESSAGE_TIMEOUT=1500; // uS to wait for a new MQ Message (or at least the initial Signature)
const unsigned long MESSAGE_TIMEOUT_EXT=6000; // EXTended MQ Message timeout, when no other actions are pending (netPending/replyPending)

// PORT PIN definitions
const byte NET_IN    =0;  // PORTD-0 on Teensy ++2.0 (Brown wire on Cypress Logic Analyser)
const byte NET_OUT   =7;  // PORTD-7 (Purple)

const byte DEBUG_NET1=1;  // PORTD-1 (Red)
const byte DEBUG_NET2=2;  // PORTD-2 (Orange)
const byte DEBUG_NET3=3;  // PORTD-3 (Yellow)
const byte DEBUG_NET4=4;  // PORTD-4 (Green)
const byte DEBUG_NET5=5;  // PORTD-5 (Blue)

const byte DEBUG_NET6=27; // PORTB-7 (White)

const byte LED_PIN_NET_RX=6; // LED pin (D3) - needs an additional LED fitted
const byte LED_PIN_NET_TX=6; // LED pin (D4) - needs an additional LED fitted
const byte LED_PIN_USB_RX=6; // LED pin (D5) - needs an additional LED fitted
const byte LED_PIN_USB_TX=6; // LED pin (D6) - built-in LED

const byte MAX_LED=4; // Count of LEDs in use
const unsigned int LED_FLASH_DURATION=63000; // 63 mS duty-period

// QLAN SEND Bit timings (uS)
const unsigned int BIT_TIME_SEND=14;         // 11uS per bit-cell
const unsigned int BIT0_TIME_SEND=20;        // 11 + 11/2 (ish) to centre of Bit-0

// QLAN RECV Bit timings (uS)
const unsigned int BIT_TIME_RECV=13;         // 11uS per bit-cell
const unsigned int BIT0_TIME_RECV=20;        // 11 + 11/2 (ish) to centre of Bit-0

// ZXNET SEND Bit timings (uS) - Works, though still room for improvement
//const unsigned int BIT_TIME_SEND=15;         // 11uS per bit-cell
//const unsigned int BIT0_TIME_SEND=20;        // 11 + 11/2 (ish) to centre of Bit-0

// ZXNET RECV Bit timings
//const unsigned int BIT_TIME_RECV=13;         // 11uS per bit-cell
//const unsigned int BIT0_TIME_RECV=22;        // 11 + 11/2 (ish) to centre of Bit-0

// v1x9b
const unsigned int DEBUG_STROBE_TIME=3;        // 3us

const byte SCOUT_BITS=16;                 // Number of bits in SCOUT
const unsigned int SCOUT_BIT_TIME=38;     // 30uS per bit-cell during (TK2 compatible) SCOUT
const unsigned int SCOUT_BIT_TIME_DIV5=7; // 30/5 us while SCOUT_BIT being checked for contention
const unsigned int SCOUT_DETECT_TIME=50;  // 50us minimum continuous ACTive before SCOUT is recognised
const unsigned int SCOUT_SKIP_TIME=475;   // 525uS skip-time over a detected SCOUT (following the initial 50uS of ACTive)

// v2.1b - Broadcast SCOUT Extention handling added

const unsigned int SCOUT_BCAST_EXT_TIME=5000;   // 5mS SCOUT (TK2) Extention-time

const unsigned int SCOUT_WORD=0b1010101010101011; // 16x Bit SCOUT, with 2x START bits and 1x STOP (positive logic-levels, LSB first)

const unsigned int LEAD_WAIT_TIME=2500;   // 2.5ms maximum wait for a H- or D- ACKnowledge or otherwise ACTive LEAD bits
const unsigned int LEAD_WAIT_TIME_2=500;  // 500us maximum wait for ACTive LEAD bits on subsequent bytes being read
const unsigned int START_WAIT_TIME=32000; // 32ms maximum wait INACTive START bit, following LEAD bit detection (Tk2 will happily wait up to 260 ms!)
// NB - the generous START_WAIT_TIME above takes TK2's SCOUT Extension (5ms) in its stride, without special consideration

const unsigned int PRE_ACK_DELAY=50;      // 50us Delay following receipt of HDR or DATA before sending H/DACK
const unsigned int POST_ACK_DELAY=125;    // 125us Delay following HACK before starting to send DATA packet
const byte VALID_ACK=0x01;                // ACKnowledge byte (currenlty only 0x01 is considered valid)

// v2.1b - Broadcast Timing Constants
const unsigned int BCAST_ACK_DETECT_TIME=625;  // 500us wait for Broadcact ACK (ndt_back)
const unsigned int BCAST_ACK_EXT_TIME=24000;   // 20ms wait during Broadcact ACK check (ndt_bace)
const unsigned int BCAST_NACK_DETECT_TIME=625; // 500us wait for Broadcact NACK (also ndt_back)

const unsigned int BCAST_ACK_TIME=5000;        // 5ms Broadcast ACK length (ndt_xack)
const unsigned int BCAST_NACK_TIME=5000;       // 5ms Broadcast NACK length (ndt_xack)
const unsigned int BCAST_NACK_WAIT=200;        // 200us wait for Broadcast ACK

// QLAN Bit lengths
const byte LEAD_BITS=2;  // Num of bits in LEADer of Byte-frame
const byte START_BITS=1; // Num of START bits of Byte-frame
const byte STOP_BITS=2;  // Num of STOP bits of Byte-frame
const byte FILL_BITS=1;  // Num of bits FILLing between consecutive Byte-framnes

// State Machine CONSTants - not yet implemented
// const int PH_GAP=0;    // PHase 0: GAP
// const int PH_SCOUT=1;  // PHase 1: SCOUT
// const int PH_HEADER=2; // PHase 2: HEADER
// const int PH_HACK=3;   // PHase 3: Header ACKnowledge
// const int PH_DATA=4;   // PHase 4: DATA
// const int PH_DACK=5;   // PHase 5: Data ACKnowledge

// const int inactCont=0;  // Level/Edge: INACTive CONTinuous
// const int activeCont=1; // Level/Edge: ACTive CONTinuous
// const int activeEdge=2; // Level/Edge: ACTive EDGE
// const int inactEdge=3;  // Level/Edge: INACTive EDGE
// const int ignore=4;     // Level/Edge: IGNORE/skip

// Timings for each PHase: { [PH, Min Duration, Max Timeout] }
// const byte PH_TIMINGS_GAP[] = { inactCont,300,2000 };
// const byte PH_TIMINGS_SCOUT[] = { activeEdge,-1,3000, activeCont,50,-1, ignore,-1,480 };
// const byte PH_TIMINGS_HEADER[] = { inactEdge,-1,6000, inactCont,11,-1 };

// Constants related to the mQTable
const int MESSAGE_HLEN=10;     // Length of each Message Header
const int MAX_MESSAGES=7;      // Maximum number of pending Messages - 1
const int MAX_MESSAGE_ID=8191; // Maximum MID before roll-over (back to 1, skipping zero)
const int MQTABLE_ENTRY_SZ=MESSAGE_HLEN + 2; // Size of each Message Entry in the mQTable=Header-length + Buffer Ptr (2xbytes)

const char MQ_SYNC[]= "%M$Q";
const int  MQ_SYNC_LEN=4;
const unsigned int MQ_SYNC_SUM=231; // Byte-additive int value of MQ_SYNC

// Message Header offsets
const int M_ID=0;
const int M_DEV=2;
const int M_ACT=3;
const int M_PARERR=4;
const int M_LENPAR2=6;
const int M_CHKSUM=8;

const int M_BUFPTR=10;

//Recognised DEVice IDs
const byte M_DEV_BRIDGE=0;  // DEVice ID for the Bridge Adapter itself
const byte M_DEV_RMS=1;     // DEVice ID for RMS (not yet implemented)
const byte M_DEV_NET=2;     // DEVice ID for NET
const byte M_DEV_MDV1=3;    // DEVice ID for MDV1_ (not yet implemented)
const byte M_DEV_MDV2=4;    // DEVice ID for MDV2_ (not yet implemented)

// Recognised ACTion Codes
const byte M_ACT_BRIDGE_RESET=0;      // Master RESET ACTion Code
const byte M_ACT_BRIDGE_STATUS=1;     // Request STATUS ACTion Code
const byte M_ACT_NETRECV=2;           // RECEIVE (NET) ACTion Code
const byte M_ACT_NETSEND=3;           // SEND (NET) ACTion Code
const byte M_ACT_NETSERV_START=6;     // FSERVe START (NET) ACTion Code
const byte M_ACT_NETSERV_STOP=7 ;     // FSERVe STOP (NET) ACTion Code
const byte M_INVALID_CMD=255;         // Unknown ACTion code/CMD
const byte M_USBERR_BAD_MQHDR=128;    // MQ Header Sync Error during USB receive
const byte M_USBERR_BAD_CHKSUM=129;   // MQ Header ChkSum Error during USB receive

const unsigned int MQ_ERROR_MASK = 0b1000000000000000; // Mask ERROR (bit 15) from M_ID (use &)
const unsigned int MQ_LONG_MASK  = 0b0100000000000000; // Mask LONG  (bit 14) from M_ID (use &)
const unsigned int MQ_REPLY_MASK = 0b0010000000000000; // Mask REPLY (bit 13) from M_ID (use &)

const unsigned int MAX_UINT = 65535;

// QDOS ERRor codes, as UINTs
const unsigned int MQ_ERR_NC     = 0b1111111111111111; // -1 : Not Complete - repeated NETwork error, cancelled
const unsigned int MQ_ERR_ORNG   = 0b1111111111111100; // -4 : Out of RaNGe - Unknown ACTion code/CMD
const unsigned int MQ_ERR_BFFL   = 0b1111111111111011; // -5 : BuFfer FuLl - No available space in SEND/RECV buffers
const unsigned int MQ_ERR_TRNS   = 0b1111111111110011; // -13: TRaNSmission error - Missing SYNC or Checksum error in USB data stream
const unsigned int MQ_ERR_IPAR   = 0b1111111111110001; // -15: Invalid PARameter - Invalid mAct or missing Message body/payload

// NET Packet sizes
const unsigned int NET_MAX_PACKET_SZ=1020; // Maximum QLAN packet size (255 x 4 for Server/Large packet)
const unsigned int NET_PACKET_SZ=255;      // Reglar (max) packet size
const unsigned int NET_HEADER_SZ=8;

// BRIdge STATUS template (char-array)
const signed int MQ_STATUS_TEXT_MAX_LEN=400;
const char *MQ_STATUS_TEXT_GEN[] = { "QLAN-USB Bridge STATUS Reply",
                                     "Firmware version",
                                     "Messages processed",
                                     "MQ Slots Total/Free",
                                     "Largest HEAP Free Space",
                                     "USB Bytes Sent/Rcvd (Drpd)",
                                     "NET packets Sent/Rcvd",
                                     "NET Bytes Sent/Rcvd",
                                     "FSERVE Listening state",
                                     "DEBUG" };                       // STATUS Reply text

// As we are feeding NET_OUT through a PNP transitor to match the QLAN hardware,
//  we need to invert the usual meaning of ACTive and INACTive
// NET_In on the otherhand is positive logic - again, matching the QLNet H/W

const boolean NET_ACTive=LOW;
const boolean NET_INACTive=HIGH;

const byte RECV_OK=0;         // Packet received and chksums matched - ACK and Queue a REPLY
const byte RECV_TRY_AGAIN=1;  // Some problem with RECV, but OK to retry - No ACK
const byte RECV_UNEXPECTED=2; // Packet HDR received OK, but did not match any pending receive - No ACK
const byte RECV_FAIL=4;       // Packet received but either the header/data chksums didn't match, or some framing error
const byte RECV_TIMEOUT=5;    // No packet detected in timeOut us

const byte HDR_GOOD=0;          // Expected Src/Dest and Block Num - ChkSum
const byte HDR_CHKSUM_ERROR=1;  // ChkSum mismatch....
const byte HDR_NOT_FOR_ME=2;    // Dst <> SELF
const byte HDR_PHANTOM_BLOCK=3; // Expected Src/Dst and ChkSum, but Block-num = (Expected - 1)
const byte HDR_BLK_MISMATCH=4;  // Block-num some other value (neither Expected, nor Expected-1)

const byte SEND_OK=0;
const byte SEND_CONTENTION=1;   //Detceted during SCOUT:InActive
const byte SEND_NO_HACK=2;      // No ACK detected in time following HDR
const byte SEND_NO_DACK=3;      // No ACK detected in time following DATA

// QLAN Header definition
const byte ND_DEST=0;
const byte ND_SELF=1;
const byte ND_BLKL=2;
const byte ND_BLKH=3;
const byte ND_TYPE=4;
const byte ND_NBYT=5;
const byte ND_DCHK=6;
const byte ND_HCHK=7;

const byte NF_SBLK=254; // FSERVE Large Block (send) ND_TYPE code

// Global Variables...


unsigned int lastGoodMId;
unsigned int lastMQError;

// unsigned int debugTiming[debugTimingEntries];
// byte nextDebugTimingEntry=0;

unsigned long ledStateArr[MAX_LED * 3] = { 3, 0, 0, 4, 0, 0, 5, 0, 0, 6, 0, 0 }; // LED State table 4x { LED_PIN, LED_STATE, LED_TIMER }

byte mQTable[MAX_MESSAGES * MQTABLE_ENTRY_SZ]; // mQTable
unsigned int mQAvailableSlots;

byte netHdrBuffer[NET_HEADER_SZ]; // Buffer for received NET packet header
// char usbSendBuffer[MESSAGE_HLEN + NET_MAX_PACKET_SZ + NET_HEADER_SZ]; // USB Send Buffer - to USB (from NET)
// char usbRecvBuffer[MESSAGE_HLEN + NET_MAX_PACKET_SZ + NET_HEADER_SZ]; // USB Receive Buffer - from USB (to NET)

// int recvPendingQueue[MAX_MESSAGES]; // Recv Pending Queue {mQSlot }

byte myStationNumber, remoteStationNumber, cmd, netRecvPendCnt;
int nextFreeMQSlot, nextReply, nextNetSend, lastNetRcvd, fServeMQSlot;
boolean mQSlotAvailable, mQScanRequired;
// boolean usbSendBufferFree, usbRecvBufferFree;
boolean netActionPending, netSendPending, netRecvPending, replyPending;

boolean fServeListening;
char    fServeListeningState[6]; // Will hold the "TRUE" or "FALSE" text-flag

char debugStr[64];

// unsigned int usbSendBuffersFree, usbRecvBuffersFree;

unsigned long messagesProcessed, usbBytesRcvd, usbBytesDropped, usbBytesSent;
unsigned long netBytesRcvd, netBytesSent, netPacketsRcvd, netPacketsSent;

// Heap Handling array and variables
byte heap[HEAP_SIZE];

void setup() {
  // put your setup code here, to run once:

  Serial.begin(115200); // For Arduino, needs to match the host PC baud, for Teensy, effectively ignored

  pinMode(NET_IN,      INPUT);
  pinMode(NET_OUT,    OUTPUT);
  pinMode(DEBUG_NET1, OUTPUT);
  pinMode(DEBUG_NET2, OUTPUT);
  pinMode(DEBUG_NET3, OUTPUT);
  pinMode(DEBUG_NET4, OUTPUT);
  pinMode(DEBUG_NET5, OUTPUT);
  pinMode(DEBUG_NET6, OUTPUT);

  pinMode(LED_PIN_NET_RX, OUTPUT);
  pinMode(LED_PIN_NET_TX, OUTPUT);
  pinMode(LED_PIN_USB_RX, OUTPUT);
  pinMode(LED_PIN_USB_TX, OUTPUT);

  digitalWrite(NET_OUT, NET_INACTive); // NB Actually HIGH to suit PNP transitor
  digitalWrite(DEBUG_NET1, LOW);       // NET DEBUG port #1
  digitalWrite(DEBUG_NET2, LOW);       // NET DEBUG port #2
  digitalWrite(DEBUG_NET3, LOW);       // NET DEBUG port #3
  digitalWrite(DEBUG_NET4, LOW);       // NET DEBUG port #4
  digitalWrite(DEBUG_NET5, LOW);       // NET DEBUG port #5
  digitalWrite(DEBUG_NET6, LOW);       // NET DEBUG port #6
  digitalWrite(LED_PIN_NET_RX, LOW);
  digitalWrite(LED_PIN_NET_TX, LOW);
  digitalWrite(LED_PIN_USB_RX, LOW);
  digitalWrite(LED_PIN_USB_TX, LOW);

}

void initialiseVariables() {

  lastGoodMId=0;
  lastMQError=0;

  // Initialise (clear) Message Queue (array)
  for (unsigned int messageIndex=0 ; messageIndex <= MAX_MESSAGES ; messageIndex++ ) {
    for (int byteIndex=0 ; byteIndex < MQTABLE_ENTRY_SZ ; byteIndex++) {
      mQTable[messageIndex * MQTABLE_ENTRY_SZ + byteIndex]=0;
    }
  }

  // Initialise (clear) Receive and Send Buffers (arrays)
//  for (unsigned int byteIndex=0 ; byteIndex < NET_MAX_PACKET_SZ + NET_HEADER_SZ ; byteIndex++ ) {
//    usbSendBuffer[byteIndex]=0;
//    usbRecvBuffer[byteIndex]=0;
//  } // for...

  // Initialise (clear) dynamic parts of ledStateArr { x, state, timer }
  for (unsigned int index=0 ; index < MAX_LED; index++ ) {
    ledStateArr[index * 3 + 1]=0;
    ledStateArr[index * 3 + 2]=0;
  } // for...

  replyPending=false;

  netSendPending=false;

  netRecvPending=false;
  netRecvPendCnt=0;
  netActionPending=false;

  nextReply=-1;
  nextNetSend=-1;

  nextFreeMQSlot=0;
  mQSlotAvailable=true;
  mQAvailableSlots=MAX_MESSAGES + 1;
  mQScanRequired=false;

//  usbSendBufferFree=true;
//  usbSendBuffersFree=1;
//  usbRecvBufferFree=true;
//  usbRecvBuffersFree=1;

  cmd=255;

  myStationNumber=1;
  remoteStationNumber=255;

  usbBytesRcvd=0;
  usbBytesDropped=0;
  usbBytesSent=0;
  messagesProcessed=0;
  fServeListening=false;
  fServeMQSlot=-1;
  strncpy(fServeListeningState, "FALSE", 5);

  strncpy(debugStr, "<none>", 6);

  netBytesRcvd=0;
  netBytesSent=0;
  netPacketsRcvd=0;
  netPacketsSent=0;

  // Initialise the HEAP
  setupHeap();

} // initialiseVariables()...

void flashLEDs(int iterations, int milSecs) {

  for (int flashes=0; flashes < iterations ; flashes++) {
    digitalWrite(LED_PIN_NET_RX, HIGH);
    digitalWrite(LED_PIN_NET_TX, HIGH);
    digitalWrite(LED_PIN_USB_RX, HIGH);
    digitalWrite(LED_PIN_USB_TX, HIGH);
    delay(milSecs);
    digitalWrite(LED_PIN_NET_RX, LOW);
    digitalWrite(LED_PIN_NET_TX, LOW);
    digitalWrite(LED_PIN_USB_RX, LOW);
    digitalWrite(LED_PIN_USB_TX, LOW);
    delay(milSecs);
  } // for...

} // flashLEDs()...

void flashLED_async() {

  byte thisLEDOffset;
  unsigned long thisLEDState, thisLEDPin, thisLEDTimer, thisLEDDuration;

  for (byte led=0; led < MAX_LED ; led++) {

    thisLEDOffset = led * 3;
    thisLEDState = ledStateArr[thisLEDOffset + 1];
    if ( thisLEDState == 0 ) continue; // Next LED, if not currently active

    thisLEDPin   = ledStateArr[thisLEDOffset];
    thisLEDTimer = ledStateArr[thisLEDOffset + 2];

    thisLEDDuration = ledTimer - thisLEDTimer;

    switch (thisLEDState) {

      case 1: // LED -ON- (already set by flashLED_start)

        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 2: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 3: // LED -ON-

        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 4: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 5: // LED -ON-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 6: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 7: // LED -ON-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = 0;
          ledStateArr[thisLEDOffset + 1] = 0; // Reset State to zero, with LED -OFF-
        } // if...
        break;

    } // switch...

  } // for...

} // flashLED_async()...

void flashLED_start(byte ledPin) {

  boolean ledPinFound=false;
  byte led, thisLEDOffset;

  for ( led=0; led < MAX_LED; led++ ) {

    if ( ledStateArr[led * 3] == long(ledPin) ) {

      ledPinFound=true;
      thisLEDOffset = led * 3;
      break;

    } // if...

  } // for...

  if ( !ledPinFound ) return;

  digitalWrite(ledPin, HIGH);
  ledStateArr[thisLEDOffset + 1] = 1;
  ledStateArr[thisLEDOffset + 2] = ledTimer;

} // flashLED_start()...

// HEAP Functions/Procedures

signed int getHeapSignedWord(signed int heapPtr) {

  unsigned int heapWordUnsigned;
  signed int heapWordSigned;

  heapWordUnsigned=word(heap[heapPtr], heap[heapPtr+1]);

  if (heapWordUnsigned>32767) {

    heapWordSigned=int(heapWordUnsigned-65536);

  } else {

    heapWordSigned=int(heapWordUnsigned);

  } // if

  return heapWordSigned;

} // getHeapSignedWord

void putHeapSignedWord(signed int heapPtr, signed int heapWordSigned) {

  unsigned int heapWordUnsigned;

  if (heapWordSigned < 0) {

    heapWordUnsigned=word(heapWordSigned+65536);

  } else {

    heapWordUnsigned=word(heapWordSigned);

  }

  heap[heapPtr]=highByte(heapWordUnsigned);
  heap[heapPtr+1]=lowByte(heapWordUnsigned);

  return;

} // putHeapSignedWord

void setupHeap() {

  for (signed int heapIndex=0; heapIndex<HEAP_SIZE; heapIndex++) heap[heapIndex]=255;

  putHeapSignedWord(HEAP_FREE_SPC_PTR,4);

  putHeapSignedWord(2,HEAP_SIZE-2);
  putHeapSignedWord(4,HEAP_END);

  return;

} // setupHeap

signed int getMaxHeapSpc() {

  signed int nextFreeSpcPtr,nextFreeSpcPtrOS,nextFreeSpcSz;
  signed int maxFreeSpcSz=0;

// return HEAP_SIZE;

  nextFreeSpcPtr=HEAP_FREE_SPC_PTR;

  // Find the maximum free-space node (if any!)
  do {

    nextFreeSpcPtrOS=getHeapSignedWord(nextFreeSpcPtr);
    if (!(nextFreeSpcPtrOS==HEAP_END)) {

      nextFreeSpcPtr+=nextFreeSpcPtrOS;
      nextFreeSpcSz=getHeapSignedWord(nextFreeSpcPtr-2);

      if (nextFreeSpcSz > maxFreeSpcSz) maxFreeSpcSz=nextFreeSpcSz;

    } // if

  } while (!(nextFreeSpcPtrOS==HEAP_END)); // do..while

  return maxFreeSpcSz-HEAP_HDR_SZ;

} // getMaxHeapSpc()

signed int allocateHeap(signed int allocSz) {

  boolean freeSpcFound;

  signed int newAllocSpcPtr,allocSzActual,nextFreeSpcPtr,prevFreeSpcPtr;
  signed int nextFreeSpcPtrOS,nextFreeSpcSz,newFreeSpcPtr,newFreeSpcSz,nodePtr;
  signed int chars;

  newAllocSpcPtr=0;
  freeSpcFound=false;
  allocSzActual=HEAP_HDR_SZ + ((allocSz+3)>>2) * 4; // Round-up to next 4-byte-boundary and add HEAP_HDR_SZ (4)

  nextFreeSpcPtr=HEAP_FREE_SPC_PTR;

  // Find a suitably sized free-space
  do {

    nextFreeSpcPtrOS=getHeapSignedWord(nextFreeSpcPtr);
    if (!(nextFreeSpcPtrOS==HEAP_END)) {

      prevFreeSpcPtr=nextFreeSpcPtr;
      nextFreeSpcPtr+=nextFreeSpcPtrOS;

      nextFreeSpcSz=getHeapSignedWord(nextFreeSpcPtr-2);
      if (nextFreeSpcSz >= allocSzActual) freeSpcFound=true;

    } // if

  } while (!(nextFreeSpcPtrOS==HEAP_END) && !freeSpcFound); // do..while

  if (!freeSpcFound) return HEAP_ERR_IMEM;

  newAllocSpcPtr=nextFreeSpcPtr;

  nodePtr=unlinkNode(nextFreeSpcPtr,prevFreeSpcPtr);

  if ((nextFreeSpcSz - (allocSzActual + 8)) < 0) {

    allocSzActual=nextFreeSpcSz;

  } else {

    newFreeSpcPtr=nextFreeSpcPtr + allocSzActual;
    nodePtr=linkNode(newFreeSpcPtr,prevFreeSpcPtr);

    newFreeSpcSz=nextFreeSpcSz - allocSzActual;
    putHeapSignedWord(newFreeSpcPtr - 2,newFreeSpcSz);

  } // if

  putHeapSignedWord(newAllocSpcPtr - 2,allocSzActual);
  putHeapSignedWord(newAllocSpcPtr,-1);

  return newAllocSpcPtr+2; // Base of USABLE space in new allocation - always positive

} // allocateHeap

signed int releaseHeap(signed int nodeBase) {

  signed int prevFreeSpcPtr,prevFreeSpcPtrOS,prevFreeSpcSz,nextFreeSpcPtr,nextFreeSpcPtrOS,nextFreeSpcSz;
  signed int newFreeSpcPtr,nodePtr,nodeSz,nodeFlag;
  boolean nodeAtStart,nodeAtEnd;

  nodePtr=nodeBase-2;
  nodeSz=getHeapSignedWord(nodePtr-2);
  nextFreeSpcPtr=HEAP_FREE_SPC_PTR;
  newFreeSpcPtr=nodePtr;

  do {

    nextFreeSpcPtrOS=getHeapSignedWord(nextFreeSpcPtr);

    prevFreeSpcPtr=nextFreeSpcPtr;
    nextFreeSpcPtr+=nextFreeSpcPtrOS;

  } while (!(nextFreeSpcPtrOS==HEAP_END) && !(nextFreeSpcPtr>nodePtr)); // do...while

  nodeAtStart=(prevFreeSpcPtr==HEAP_FREE_SPC_PTR);
  nodeAtEnd=(nextFreeSpcPtrOS==HEAP_END);


// This commented-out section would suffice very nicely, were we not also attempting to merge contiguous free-space nodes
//  if (nodeAtEnd==false) {

//      putHeapSignedWord(prevFreeSpcPtr,(nodePtr-prevFreeSpcPtr));
//      putHeapSignedWord(nodePtr,(nextFreeSpcPtr-nodePtr));

//  } else {

//      putHeapSignedWord(nextFreeSpcPtr, (nodePtr-nextFreeSpcPtr));
//      putHeapSignedWord(nodePtr,HEAP_END);

//  } // if

// Update pointers between the nodes, whilst checking for contiguous free-space nodes and merge them accordingly
//   a) prev>node>next: resulting in a single new node
//   b) prev>node, next: 2 nodes
//   c) prev, node>next: 2 nodes
//   d) No continuity: 3 nodes

  nodeFlag=int(nodeAtStart==true)+2*(int(nodeAtEnd==true));

  switch (nodeFlag) {

    case 0:
    // node falls between two existing free-space nodes (prev < node < next): neither nodeAtStart nor nodeAtEnd

      prevFreeSpcSz=getHeapSignedWord(prevFreeSpcPtr-2);
      prevFreeSpcPtrOS=nextFreeSpcPtrOS;

      nextFreeSpcSz=getHeapSignedWord(nextFreeSpcPtr-2);
      nextFreeSpcPtrOS=getHeapSignedWord(nextFreeSpcPtr);

      if ((prevFreeSpcSz+nodeSz)==(nextFreeSpcPtr-prevFreeSpcPtr)) {
      // Merge node with free-space nodes on either side, creating a single new node...

        putHeapSignedWord(prevFreeSpcPtr-2,(prevFreeSpcSz+nodeSz+nextFreeSpcSz));

        if (nextFreeSpcPtrOS==HEAP_END) {

          putHeapSignedWord(prevFreeSpcPtr,HEAP_END);

        } else {

          putHeapSignedWord(prevFreeSpcPtr,(prevFreeSpcPtrOS+nextFreeSpcPtrOS));

        } // if

        newFreeSpcPtr=prevFreeSpcPtr;

      } else if (prevFreeSpcSz==(nodePtr-prevFreeSpcPtr)) {
      // Merge node with previous free-space node, resulting in two nodes...

        putHeapSignedWord(prevFreeSpcPtr-2,(prevFreeSpcSz+nodeSz));

        newFreeSpcPtr=prevFreeSpcPtr;

      } else if (nodeSz==(nextFreeSpcPtr-nodePtr)) {
      // Merge node with next free-space node, resulting in two nodes...

        putHeapSignedWord(prevFreeSpcPtr,(nodePtr-prevFreeSpcPtr));
        putHeapSignedWord(nodePtr-2,(nodeSz+nextFreeSpcSz));

        if (nextFreeSpcPtrOS==HEAP_END) {

          putHeapSignedWord(nodePtr,HEAP_END);

        } else {

          putHeapSignedWord(nodePtr,(nodeSz+nextFreeSpcPtrOS));

        } // if

      } else {
      // No merging required...

        putHeapSignedWord(prevFreeSpcPtr,(nodePtr-prevFreeSpcPtr));
        putHeapSignedWord(nodePtr,(nextFreeSpcPtr-nodePtr));

      } // if

      break;

    case 1:
    // node falls before first free-space node (node > prev=FSP, node < next): nodeAtStart

      nextFreeSpcSz=getHeapSignedWord(nextFreeSpcPtr-2);
      // prevFreeSpcSz=getHeapSignedWord(prevFreeSpcPtr-2);

      if (nodeSz==(nextFreeSpcPtr-nodePtr)) {
      // Merge node with next free-space node...

        nextFreeSpcPtrOS=getHeapSignedWord(nextFreeSpcPtr);

        putHeapSignedWord(prevFreeSpcPtr,(nodePtr-prevFreeSpcPtr));
        putHeapSignedWord(nodePtr-2,(nodeSz+nextFreeSpcSz));

        if (nextFreeSpcPtrOS==HEAP_END) {

          putHeapSignedWord(nodePtr,HEAP_END);

        } else {

          putHeapSignedWord(nodePtr,(nodeSz+nextFreeSpcPtrOS));

        } // if

      } else {
      // No merging required...

        putHeapSignedWord(prevFreeSpcPtr,(nodePtr-prevFreeSpcPtr));
        putHeapSignedWord(nodePtr,(nextFreeSpcPtr-nodePtr));

      } // if

      break;

    case 2:
    // node falls after last free-space node (node > prev=next): nodeAtEnd

      nextFreeSpcSz=getHeapSignedWord(nextFreeSpcPtr-2);
      // prevFreeSpcSz=getHeapSignedWord(prevFreeSpcPtr-2);

      if (nextFreeSpcSz==(nodePtr-nextFreeSpcPtr)) {
      // Merge node with previous free-space node...

        putHeapSignedWord(nextFreeSpcPtr-2,(nextFreeSpcSz+nodeSz));
        // putHeapSignedWord(nextFreeSpcPtr,HEAP_END);

        newFreeSpcPtr=nextFreeSpcPtr;

      } else {
      // No merging required...

        putHeapSignedWord(nextFreeSpcPtr,(nodePtr-nextFreeSpcPtr));
        putHeapSignedWord(nodePtr,HEAP_END);

      } // if

      break;

    case 3:
    // heap had no free-space (node > prev=next=FSP): nodeAtEnd/nodeAtStart
    // No merging required...

//      putHeapSignedWord(nextFreeSpcPtr, (nodePtr-nextFreeSpcPtr));
      putHeapSignedWord(HEAP_FREE_SPC_PTR, (nodePtr-HEAP_FREE_SPC_PTR));
      putHeapSignedWord(nodePtr,HEAP_END);

      break;

  } // switch

  return newFreeSpcPtr+2;

} // releaseHeap

signed int linkNode(signed int nodePtr,signed int prevPtr) {

 signed int prevPtrOS,nextPtr,newPtrOS;

  prevPtrOS=getHeapSignedWord(prevPtr);

  if (prevPtrOS==HEAP_END) {

    newPtrOS=HEAP_END;

  } else {

    nextPtr=prevPtr+prevPtrOS;
    newPtrOS=nextPtr-nodePtr;

  } // if

  prevPtrOS=nodePtr-prevPtr;

  putHeapSignedWord(prevPtr,prevPtrOS);
  putHeapSignedWord(nodePtr,newPtrOS);

  return prevPtr;

} // linkNode

signed int unlinkNode(signed int nodePtr,signed int startPtr) {

  signed int nextPtr,newPtrOS,prevPtr,nodePtrOS,nextPtrOS;

  nextPtr=startPtr;

  do {

    nextPtrOS=getHeapSignedWord(nextPtr);
    prevPtr=nextPtr;
    nextPtr+=nextPtrOS;

  } while (!(nextPtrOS==HEAP_END) && !(nextPtr==nodePtr)); // do...while

  if (nextPtrOS==HEAP_END) {

    newPtrOS=HEAP_END;

  } else {

    nodePtrOS=getHeapSignedWord(nodePtr);

    if (nodePtrOS==HEAP_END) {

      newPtrOS=HEAP_END;

    } else {

      nextPtr=nodePtr+nodePtrOS;
      newPtrOS=nextPtr - prevPtr;

    } // if

  } // if

  putHeapSignedWord(prevPtr,newPtrOS);

  return prevPtr;

} // unlinkNode

void putMQ(int mQSlot, unsigned int mID, byte mDev, byte mAct, unsigned int mParErr, unsigned int mLenPar2, unsigned int mChkSum) {
  // Update Message Queue at mQSlot

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;
  mQTable[mQOffset]                 = highByte(mID);
  mQTable[mQOffset + 1]             = lowByte(mID);
  mQTable[mQOffset + M_DEV]         = mDev;
  mQTable[mQOffset + M_ACT]         = mAct;
  mQTable[mQOffset + M_PARERR]      = highByte(mParErr);
  mQTable[mQOffset + M_PARERR + 1]  = lowByte(mParErr);
  mQTable[mQOffset + M_LENPAR2]     = highByte(mLenPar2);
  mQTable[mQOffset + M_LENPAR2 + 1] = lowByte(mLenPar2);
  mQTable[mQOffset + M_CHKSUM]      = highByte(mChkSum);
  mQTable[mQOffset + M_CHKSUM + 1]  = lowByte(mChkSum);
  // mQTable[mQOffset + M_BUFPTR]      = highByte(mBufPtr);
  // mQTable[mQOffset + M_BUFPTR + 1]  = lowByte(mBufPtr);

} // putMQ...

// v2.1b - Added to support updating MParErr indepedently (for 'Listen to Self' type Net CMDs)
void putMQParErr(int mQSlot, unsigned int mParErr) {
  // Update Message Queue at mQSlot with updated M_PARERR

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;
  mQTable[mQOffset + M_PARERR]      = highByte(mParErr);
  mQTable[mQOffset + M_PARERR + 1]  = lowByte(mParErr);

} // putMQParErr...

void putMQBufPtr(int mQSlot, signed int mBufPtr) {
  // Update Message Queue at mQSlot with M_BUFPTR

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;
  mQTable[mQOffset + M_BUFPTR]      = highByte(mBufPtr);
  mQTable[mQOffset + M_BUFPTR + 1]  = lowByte(mBufPtr);

} // putMQBufPtr...

void markMQSlotEmpty(int mQSlot) {
  // Update Message Queue at mQSlot with mID=0;

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;

  for ( byte byteIndex=0; byteIndex < MQTABLE_ENTRY_SZ; byteIndex++ ) mQTable[mQOffset + byteIndex] = 0;

} // markMQSlotEmpty()...

void getMQ(int mQSlot, unsigned int &mID, byte &mDev, byte &mAct, unsigned int &mParErr, unsigned int &mLenPar2, unsigned int &mChkSum) {
  /* Read-back Message Queue details from mQSlot
   * NB - Use of simple dereferencing '&' to allow passing-back values without resorting to pointers.
  */

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;

  mID      = word( mQTable[mQOffset], mQTable[mQOffset + 1] );
  mDev     = mQTable[mQOffset + M_DEV];
  mAct     = mQTable[mQOffset + M_ACT];
  mParErr  = word( mQTable[mQOffset + M_PARERR],  mQTable[mQOffset + M_PARERR + 1] );
  mLenPar2 = word( mQTable[mQOffset + M_LENPAR2], mQTable[mQOffset + M_LENPAR2 + 1] );
  mChkSum  = word( mQTable[mQOffset + M_CHKSUM],  mQTable[mQOffset + M_CHKSUM + 1] );
  // mBufPtr  = word( mQTable[mQOffset + M_BUFPTR],  mQTable[mQOffset + M_BUFPTR + 1] );

} // getMQ...

signed int getMQBufPtr(int mQSlot) {
  /* Read-back M_BUFPTR from mQSlot in Message Queue
  */

  int mQOffset;
  signed int mBufPtr;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return 0; // Out of bounds

  mQOffset = mQSlot * MQTABLE_ENTRY_SZ;
  mBufPtr  = word( mQTable[mQOffset + M_BUFPTR],  mQTable[mQOffset + M_BUFPTR + 1] );

  return mBufPtr;

} // getMQBufPtr...

unsigned int calculateMQChkSum(unsigned int mID, byte mDev, byte mAct, unsigned int mParErr, unsigned int mLenPar2) {
  // Calculate the byte-additive ChkSum of MQ_SYNC + mID + mDev + mAct + mParErr + mLenPar2

  unsigned int chkSum = MQ_SYNC_SUM; // Initialise CHKSUM to 231d

  chkSum+= highByte(mID);
  chkSum+= lowByte(mID);
  chkSum+= mDev;
  chkSum+= mAct;
  chkSum+= highByte(mParErr);
  chkSum+= lowByte(mParErr);
  chkSum+= highByte(mLenPar2);
  chkSum+= lowByte(mLenPar2);

  return chkSum;

} // calculateMQChkSum()...

void checkUSBMessage() {
  // Check USB for a new Message and post it in nextFreeMQSlot if available (mQSlotAvailable=true)

  boolean messageReceived, mQSyncFound, longMessageExpected, longMessageRcvd;
  boolean bufferError;

  int nextMQSyncIndex, mQOffset;

  byte mDev, mAct, nextMQSyncByte, nextByte;

  unsigned int mQTimeout, mID, mParErr, mLenPar2, mChkSum, newChkSum;
  signed int mBufPtr,largestHeapFreeSpc;
  unsigned int byteIndex;

  if ( !mQSlotAvailable ) {

    PORTB |= 0b10000000; // DEBUG_NET6 - 25us: MQ_ERROR
    delayMicroseconds(25);
    PORTB &= 0b01111111;

    return;

  }

  mQScanRequired=true;

  messageReceived=false;
  longMessageExpected=false;
  longMessageRcvd=false;
  bufferError=false;

  if ( netActionPending || replyPending ) {

    mQTimeout = MESSAGE_TIMEOUT; // 1.5ms

  } else {

    mQTimeout = MESSAGE_TIMEOUT_EXT;  // 6ms

  }

  timer=0; // Reset the timer - we only stay in here until a new message is received, or else until mQTimeout expires

  while ( timer < mQTimeout && !messageReceived ) {
    //  while-0: Terminate reading USB if timeout without a valid MQSync/Header being found, or else when a full message is read

    mQSyncFound=false;
    nextMQSyncIndex=0;

// 1. Check for MQSYNC

    // First, scan for MQ Sync Signature
    do {
      // do-0.1:

      if ( !Serial.available() ) continue;

// Try this... Reset the timer now that we've found a character, allowing another mQTimeout to lapse
      timer=0;

      nextMQSyncByte=byte( MQ_SYNC[nextMQSyncIndex] );

      nextByte=Serial.read();
      usbBytesRcvd++;

      if ( nextByte != nextMQSyncByte ) {

        flashLEDs(2, 1000);
        usbBytesDropped += nextMQSyncIndex;

        if ( nextByte == byte(MQ_SYNC[0]) ) {

          nextMQSyncIndex=1;
          continue; // @do-0.1: Accomodate a repeated match at the beginning of MQ_SYNC, without losing the most recent captured byte from the data-stream - takes a bit of thinking about...

        } else {   

          usbBytesDropped++; // adjust the number of bytes dropped...   
          nextMQSyncIndex=0; // Start over from beginning of MQ_SYNC, from next byte in the data-stream          

          continue; // @do-0.1

        } // if...

      } // if...

      if ( nextMQSyncIndex < (MQ_SYNC_LEN - 1) ) {

        nextMQSyncIndex++;

      } else {

        mQSyncFound=true;

      } // if...

    } while ( timer < mQTimeout && !mQSyncFound ); // do-0.1...while...

    if ( !mQSyncFound ) {

// Prepare BAD_MQSYNC Reply and return

      return; // We timed-out looking for a fresh MQ Sync Signature

    } // if...

// 2. Check for a valid MQHDR

    /* So, we've found the MQ Sync signature and we haven't yet timed-out - read MQ Header with an extended timeout...
    *   NB - it is still quite possible that the byte stream we found was NOT a real SYNC and, whilst the ChkSum check later should cause this block
    *        to be discarded, it is less likely, but possible, that the 'real' SYNC is contained in what we capture as the Header - and will thus be missed.
    *        Unless, of course, we go to some lengths to scan the header for a potential SYNC as well. Thanks, but no thanks...
    */

    mQOffset = nextFreeMQSlot * MQTABLE_ENTRY_SZ;
    byteIndex=0; // replacement for for-loop, so we can 'continue' if no bytes yet available from USB without incrementing byteIndex

    /* TODO: With a bit more thought, it should be possible to reproduce the host-side approach to incomplete messages
    *  - namely, if we timeout in this iteration, we record the byte-position we got to (whether in the HDR or a LONG body) and
    *    carry-on where we left off on the next invocation of checkUSBMessage().
    *    We should ultimately timeout, however (say, after 3 retries/invocations), so as not to block processing a new message for-ever.
    */

    do {
      // do-0.2

      if ( timer > MESSAGE_TIMEOUT_EXT ) {

        // mQScanRequired=false; // Nothing new in the MQ to talk-about...

        if ( byteIndex > 0 ) {

          markMQSlotEmpty( nextFreeMQSlot ); // Mark (partially filled) MQ Slot 'empty'

        } // if...

        usbBytesDropped += ( MQ_SYNC_LEN + byteIndex );

// Prepare BAD_MQHDR Reply and return
        mAct=M_USBERR_BAD_MQHDR;

        return; // We timed-out waiting for a (complete) Message Header

      } // if...

      if ( !Serial.available() ) continue;

      mQTable[mQOffset + byteIndex] = Serial.read();
      usbBytesRcvd++;
      byteIndex++;

    } while ( byteIndex < MESSAGE_HLEN ); // do-0.2...while...

// 3. Got MQSYNC and MQHDR - validate and action

    // OK, got the header of a new message in to mQTable at nextFreeMQSlot - lets first check ChkSum and, if there's more to read, do so

    getMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );

    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

    if ( mChkSum != newChkSum ) {

// Prepare BAD_MQCHKSUM Reply and return

      mAct=M_USBERR_BAD_CHKSUM;

      flashLEDs(4, 1000);
      usbBytesDropped += ( MQ_SYNC_LEN + MESSAGE_HLEN );
      markMQSlotEmpty( nextFreeMQSlot ); // Mark MQ Slot 'empty'

      continue; // @while-0: Check sum didn't match sync+data, continue to scan data stream until we find another MQ Sync, or timeout (what else can we do?!)

    } // if...

    // If we get here, then we've successfully received a new message HDR, with a valid checksum
    // - let's validate the M_DEV/M_ACT combinations that we support today.

    if ( ( ( mDev == M_DEV_NET && 
         ( ( mAct == M_ACT_NETRECV ) ||
      ( mAct == M_ACT_NETSEND ) ||
      ( mAct == M_ACT_NETSERV_START ) || 
      ( mAct == M_ACT_NETSERV_STOP ) ) )
        ||
        ( mDev == M_DEV_BRIDGE && 
          ( ( mAct == M_ACT_BRIDGE_RESET ) || 
      ( mAct == M_ACT_BRIDGE_STATUS ) ) ) 
      ) ) {

      cmd=mAct;

      if (cmd == M_ACT_NETSEND ) {

        longMessageExpected=true;
        if ( mLenPar2 > getMaxHeapSpc() ) bufferError=true;

      } // if...

    } else {

      cmd=M_INVALID_CMD; // Only dealing with NET:RECEIVE, SEND, SERV_START, SERV_STOP & BRIDGE:RESET, STATUS device requests today...

    } // if...

    messageReceived=true; // MQ Header looks good - read rest of message-body (if any) assuming sufficient space in HEAP
    messagesProcessed++;

    if ( mID & MQ_LONG_MASK ) {

      // LONG Message - try to capture the Message body in to HEAP if required, or else purge
      longMessageRcvd=true;

      if ( longMessageExpected && !bufferError ) {
        // Allocate, then capture to a new space in HEAP (we've already verified that a large enough space exists)

        mBufPtr=allocateHeap(mLenPar2);
        putMQBufPtr(nextFreeMQSlot,mBufPtr);

        for ( byteIndex=0; byteIndex < mLenPar2; byteIndex++ ) {
        // for-0.1

          if ( Serial.available() ) {

            heap[mBufPtr+byteIndex] = Serial.read();
            usbBytesRcvd++;

          } else {

            longMessageRcvd=false;
            mBufPtr=releaseHeap(mBufPtr);
            putMQBufPtr(nextFreeMQSlot,0);
            break; // for-0.1

          } // if...

        } // for...

//        if ( longMessageRcvd ) {

//          usbRecvBuffersFree--;
//          if ( usbRecvBuffersFree < 1 ) usbRecvBufferFree=false;

//        } // if...

      } else {
        // Can't - or don't need to - store the message body - just purge what's avaialble

        longMessageRcvd=false;

        for ( byteIndex=0; byteIndex < mLenPar2; byteIndex++ ) {
        // for-0.2

          if ( Serial.available() ) {

          nextByte=Serial.read();
          usbBytesDropped++;

          } else {

            break; // for-0.2

          } // if...           

        } // for...

      } // if...

    } // if...

// 4. Process command, or return

    if ( messageReceived ) {

      lastGoodMId=( ( ( mID & ~MQ_REPLY_MASK ) & ~MQ_LONG_MASK ) & ~MQ_ERROR_MASK );

      flashLED_start(LED_PIN_USB_RX);

      mQAvailableSlots--;
      if ( mQAvailableSlots < 1  ) mQSlotAvailable=false;

      netActionPending=false;
      replyPending=false;

      switch (cmd) {
        // switch-0.1

        case M_ACT_BRIDGE_RESET:
          // Mark the Message for REPLY and prioritise for next reply action (no further replies will then take place - due to re-initialisation)

          mID = mID | MQ_REPLY_MASK;  // Set bit-13 to mark as REPLY
          mID = mID & ~MQ_LONG_MASK;  // Clear bit-14 to mark as SHORT
          mID = mID & ~MQ_ERROR_MASK; // Clear bit-15 to mark NO ERRor

          mParErr=0; // No ERRor/OK
          mLenPar2=0;
          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

          nextReply = nextFreeMQSlot;
          netActionPending=false;
          netSendPending=false;
          netRecvPending=false;
          replyPending=true;
          mQScanRequired=false;

          // return; // No need to revisit the MQ table as we'll re-initialise immediately once the REPLY is sent...
          break; // switch-0.1

        case M_ACT_BRIDGE_STATUS:
          // Add a STATUS REPLY Message to the MQ

          if ( MQ_STATUS_TEXT_MAX_LEN > getMaxHeapSpc() ) {

            mID = mID | MQ_REPLY_MASK; // Set bit-13 to mark as REPLY
            mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark as SHORT
            mID = mID | MQ_ERROR_MASK; // Set bit-15 to mark as ERROR

            mParErr = MQ_ERR_BFFL; // ERRor Buffer Full
            mLenPar2=0;

          } else {

            mID = mID | MQ_REPLY_MASK; // Set bit-13 to mark as REPLY
            mID = mID | MQ_LONG_MASK;  // Set bit-14 to mark as LONG
            mID = mID & ~MQ_ERROR_MASK; // Clear bit-15 to mark NO ERRor

            mParErr=0; // No ERRor/OK

            /* Queue the STATUS REPLY here...
            *    QLAN-USB Bridge STATUS Reply
            *    ----------------------------
            *    Firmware version:
            *    Messages processed:
            *    MQ Slots Total/Free:
            *    Largest HEAP Free Space:
            *    USB Bytes Sent/Rcvd (Drpd):
            *    NET packets Sent/Rcvd:
            *    NET Bytes Sent/Rcvd:
            *    FSERVE Listening state:
            *    DEBUG:
            */

            largestHeapFreeSpc=getMaxHeapSpc(); // NB Max space BEFORE we allocate this STATUS reply

for (byteIndex=0; byteIndex<16; byteIndex++) {
  mLenPar2=snprintf( &debugStr[byteIndex*3], 4, "%X  ",heap[byteIndex]);
} // for

            mBufPtr=allocateHeap(MQ_STATUS_TEXT_MAX_LEN);
            putMQBufPtr(nextFreeMQSlot,mBufPtr);

            mLenPar2 = snprintf( (char*) &heap[mBufPtr], MQ_STATUS_TEXT_MAX_LEN,
                       "%s\n----------------------------\n%s: %s\n%s: %lu\n%s: %u/%u\n%s: %u\n%s: %lu/%lu (%lu)\n%s: %lu/%lu\n%s: %lu/%lu\n%s: %s\n%s: %s",
                       MQ_STATUS_TEXT_GEN[0],
                       MQ_STATUS_TEXT_GEN[1], QLAN_USB_VERSION,
                       MQ_STATUS_TEXT_GEN[2], messagesProcessed,
                       MQ_STATUS_TEXT_GEN[3], MAX_MESSAGES + 1, mQAvailableSlots,
                       MQ_STATUS_TEXT_GEN[4], largestHeapFreeSpc,
                       MQ_STATUS_TEXT_GEN[5], usbBytesSent, usbBytesRcvd, usbBytesDropped,
                       MQ_STATUS_TEXT_GEN[6], netPacketsSent, netPacketsRcvd,
                       MQ_STATUS_TEXT_GEN[7], netBytesSent, netBytesRcvd,
                       MQ_STATUS_TEXT_GEN[8], fServeListeningState,
                       MQ_STATUS_TEXT_GEN[9], debugStr
                       );

//            usbSendBuffersFree--;
//            if ( usbSendBuffersFree < 1 ) usbSendBufferFree=false;

          } // if...

          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

          nextReply = nextFreeMQSlot;
          netSendPending=false;
          netRecvPending=false;
          netActionPending=false;
          replyPending=true;
          mQScanRequired=false;

          // return; // No need to revisit the MQ table - we'll prioritise the STATUS reply
          break; // switch-0.1

        case M_ACT_NETRECV:
          /* DST.b & SRC.b in mParErr, Block.l & Block.h in mLenPar2.w
           *  - nothing to do here for now...
          */

          break; // switch-0.1

        case M_ACT_NETSEND:
          /* Expect the Packet to SEND in the message body - if available, it has been read in to HEAP at mBufPtr
           *  - otherwise REPLY with error... ERR_IPAR (-15) or ERR_BFFL (-13)
          */

          if ( !longMessageRcvd ) {
            // We expect LONG message for this CMD - if not, it's a protocol error - send an ERRor REPLY

            mID = mID | MQ_REPLY_MASK;  // Set bit-13 to mark as REPLY
            mID = mID & ~MQ_LONG_MASK;  // Clear bit-14 to mark SHORT
            mID = mID | MQ_ERROR_MASK;  // Set bit-15 to mark as ERROR

            if ( bufferError ) {

              mParErr = MQ_ERR_BFFL; // ERRor BuFfer FuLl

            } else {

              mParErr = MQ_ERR_IPAR; // ERRor Invalid PARameter

            } // if...

            mLenPar2=0;
            newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
            putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

          } // if...

          break; // switch-0.1

        case M_ACT_NETSERV_START:
          // Add a persistent FSERVE Message to the MQ - like a RECEIVE, but won't be removed after REPLYing with the received packet - no error if already set

          if ( fServeListening ) {
            // Already listening for FSERVE packets - mark this message 'empty'

            markMQSlotEmpty( nextFreeMQSlot );

          } else {

            fServeListening=true;
            strncpy(fServeListeningState, "TRUE ", 5);
            fServeMQSlot=nextFreeMQSlot;

          }

          break; // switch-0.1

        case M_ACT_NETSERV_STOP:
          // Remove persistent FSERVE Message from the MQ, if it exists (no error if not)

          markMQSlotEmpty( nextFreeMQSlot );

          if ( fServeListening ) {
            // Was listening for FSERVE packets - mark that message (fServeMQSlot) also as 'empty'

            markMQSlotEmpty( fServeMQSlot );
            fServeListening=false;
            strncpy(fServeListeningState, "FALSE", 5);
            fServeMQSlot=-1;

          }          

          break; // switch-0.1

        case M_INVALID_CMD:
          // Fall-through to default catch-all

        default:

          flashLEDs(3, 1000);

          // Unknown CMD in Message - prepare REPLY with Unknown CMD ERRor (ORNG)

          mID = mID | MQ_REPLY_MASK; // Set bit-13 to mark as REPLY
          mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark SHORT
          mID = mID | MQ_ERROR_MASK; // Set bit-15 to mark as ERRor

          mParErr=MQ_ERR_ORNG; // ERRor Out of RaNGe
          mLenPar2=0;

          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

          break; // switch-0.1

      } // switch-0.1...

    } // if...

  } // while-0...

} // checkUSBMessage()...

void scanMessageQueue() {
  // Now find next available MQ Slot (M_ID's == 0), any pending Replies or netActions...

  boolean thisReplyPending, thisNetSendPending;
  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum, lowestReplySoFar, lowestNetSendSoFar;
  unsigned int mIDMasked;

  mQSlotAvailable=false;
  mQAvailableSlots=0;

  netActionPending=false;

  netSendPending=false;

  netRecvPending=false;

  nextNetSend=-1;
  netRecvPendCnt=0;

  lowestReplySoFar=MAX_UINT;
  lowestNetSendSoFar=MAX_UINT;

  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {

    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );

    mIDMasked = mID       & ~MQ_REPLY_MASK;
    mIDMasked = mIDMasked & ~MQ_LONG_MASK;
    mIDMasked = mIDMasked & ~MQ_ERROR_MASK;

    if ( mIDMasked == 0 ) {
      // This SHOULDN'T happen...

      markMQSlotEmpty( mQSlot );
      mID = 0;

    } // if...

    if ( mID == 0 ) {

      mQAvailableSlots++;

      if ( !mQSlotAvailable ) {

        nextFreeMQSlot = mQSlot; // Find the lowest free MQ Slot
        mQSlotAvailable = true;

      } // if...

      continue; // This slot is 'empty' - next mQSlot...

    } // if...

    thisReplyPending=false;
    thisNetSendPending=false;

    if ( mID & MQ_REPLY_MASK ) thisReplyPending = true;

    if ( ( mAct == M_ACT_NETRECV || mAct == M_ACT_NETSERV_START ) && !thisReplyPending ) {

      netRecvPendCnt++;
      netRecvPending = true;

    } // if

    if ( (mAct == M_ACT_NETSEND) && !thisReplyPending ) thisNetSendPending = true;

    mID = mIDMasked; // Mask-out the REPLY, LONG & ERROR bits (temporarily - still marked as such in the MQ record)

    if ( thisReplyPending && (mID < lowestReplySoFar) ) {

      replyPending = true;
      nextReply = mQSlot;
      lowestReplySoFar = mID;
      continue; // Next mQSlot...

    } // if...

    if ( thisNetSendPending && (mID < lowestNetSendSoFar) ) {

      netSendPending = true;
      nextNetSend = mQSlot;
      lowestNetSendSoFar = mID;

    } // if...

  } // for...

  if ( netSendPending || netRecvPending ) netActionPending=true;

} // scanMessageQueue...

void actionNET() {
  /* Depending upon netSend and netRecvPending flags, listen/transmit across the wire:
   *  netSend: check for 3ms of continuous inactivity, then transmit nextNetSend message
   *  netRecv: check for at least 200us and no more than 20ms of inactivity
   *  netSend & netRecv: check for at least 200us of inactivity and not more than 3ms, then switch to transmit instead
  */

  boolean gapFound, packetReceived, packetSent;

  unsigned int minRecvGap, maxRecvGap, minSendGap;
  unsigned int mID, mParErr, mLenPar2, mChkSum, newChkSum;
  signed int mBufPtr;

  byte packetRecvCode, packetSendCode;
  byte mDev, mAct;

  minRecvGap=200;   // Minimum GAP before we'll recognise a packet for receive
  maxRecvGap=20000; // Maximum GAP before we timeout waiting for a packet
  minSendGap=3000;  // Minimum GAP before we'll start to transmit a packet

  gapFound=false;
  packetReceived=false;
  lastNetRcvd=-1;
  packetSent=false;

  packetRecvCode=-1;
  packetSendCode=-1;

  if ( netSendPending && netRecvPending ) {

    gapFound = waitGap(minRecvGap);
    if ( gapFound == false ) {

      PORTD |= 0b00010000; // DEBUG_NET4 -  5us: RECV_NO_GAP
      delayMicroseconds(5);
      PORTD &= 0b11101111; 

      return; // no minimum GAP found - just quit for now (retrying on next cycle)

    }

    packetRecvCode=listenForPacket(minSendGap - minRecvGap, lastNetRcvd);
    // NB lastNetRcvd is passed by &reference

    switch (packetRecvCode) {

      case RECV_OK:
        // Packet received OK (no chksum errors) and (started) within the time-out
        // lastNetRcvd will have been updated with the MQ Slot ID needed for the reply

        netPacketsRcvd++;
        packetReceived=true;
        break;

      case RECV_FAIL:
        // Packet detected but either chksum mismatch or some other framing error

        return;

      case RECV_UNEXPECTED:
        // Packet received successfully, but not one that we were expecting (Dst/Src & Blk-L/Blk-H)

        return;

      case RECV_TIMEOUT:
        // No packet received, let's move-on to transmit instead (minSendGap already reached...)

        packetSendCode=sendPacket(nextNetSend);
        switch (packetSendCode) {

          case SEND_OK:
            // Packet successfully sent/acknowledged

            netPacketsSent++;
//            usbRecvBuffersFree++;
//            usbRecvBufferFree=true;
            packetSent=true;
            break;

          case SEND_CONTENTION:
            // Send failed due to contention detected during SCOUT, quit for now (retrying on next cycle)

            PORTD |= 0b00100000; // DEBUG_NET5 - 15us: SEND_CONTENTION
            delayMicroseconds(15);
            PORTD &= 0b11011111;

            return;

          case SEND_NO_HACK:
            // Send failed due to missing HDR ACKnowledge from DST node

            PORTD |= 0b00100000; // DEBUG_NET5 - 25us: SEND_NO_HACK
            delayMicroseconds(25);
            PORTD &= 0b11011111;

            return;

          case SEND_NO_DACK:
            // Send failed due to missing DATA ACKnowledge from DST node

            PORTD |= 0b00100000; // DEBUG_NET5 - 35us: SEND_NO_DACK
            delayMicroseconds(35);
            PORTD &= 0b11011111;

            return;

        } // switch...

    } // switch...

  } else if ( netSendPending ) {
    // Only netSendPending

    gapFound = waitGap(minSendGap);
    if ( gapFound == false ) {

      PORTD |= 0b00100000; // DEBUG_NET5 -  5us: SEND_NO_GAP
      delayMicroseconds(5);
      PORTD &= 0b11011111;

      return;

    }

    packetSendCode=sendPacket(nextNetSend);
    switch (packetSendCode) {

      case SEND_OK:
        // Packet successfully sent/acknowledged

        netPacketsSent++;
//        usbRecvBuffersFree++;
//        usbRecvBufferFree=true;
        packetSent=true;
        break;

      case SEND_CONTENTION:
        // Send failed due to contention detected during SCOUT, quit for now (retrying on next cycle)

	PORTD |= 0b00100000; // DEBUG_NET5 - 15us: SEND_CONTENTION
	delayMicroseconds(15);
	PORTD &= 0b11011111;

        return;

      case SEND_NO_HACK:
        // Send failed due to missing HDR ACKnowledge from DST node

	PORTD |= 0b00100000; // DEBUG_NET5 - 25us: SEND_NO_HACK
	delayMicroseconds(25);
	PORTD &= 0b11011111;

        return;

      case SEND_NO_DACK:
        // Send failed due to missing DATA ACKnowledge from DST node

	PORTD |= 0b00100000; // DEBUG_NET5 - 35us: SEND_NO_DACK
	delayMicroseconds(35);
	PORTD &= 0b11011111;

        return;

    } // switch...

  } else {
  // Only netRecvPending

    gapFound = waitGap(minRecvGap);
    if ( gapFound == false ) {

      PORTD |= 0b00010000; // DEBUG_NET4 -  5us: RECV_NO_GAP
      delayMicroseconds(5);
      PORTD &= 0b11101111;

      return; // no minimum GAP found - just quit for now (retrying on next cycle)

    }

    packetRecvCode=listenForPacket(maxRecvGap - minRecvGap, lastNetRcvd);
    // NB lastNetRcvd is passed by &reference and points to the MQ slot that matched the received packet

    switch (packetRecvCode) {

      case RECV_OK:
        // Packet received OK (no chksum errors) and (started) within the time-out

        netPacketsRcvd++;
        packetReceived=true;
        break;

      case RECV_FAIL:
        // Packet detected but either chksum mismatch or some other framing error

        return;

      case RECV_UNEXPECTED:
        // Packet received successfully, but not one that we were expecting (Dst/Src & Blk-L/Blk-H)

        return;

      case RECV_TIMEOUT:
        // No (valid) packet received - just quit for now (retrying on next cycle)

        return;

    } // switch...

  } // if...

  // Now prepare any REPLY, based on packetSent or packetReceived.
  // - in practise, we shouldn't get here unless at least one is true

  if ( packetSent ) {

    getMQ( nextNetSend, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for nextNetSend
    mID = mID | MQ_REPLY_MASK; // Set bit-13 to mark as REPLY
    mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark as SHORT
    mParErr=0;
    mLenPar2=0;

    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

    putMQ( nextNetSend, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

    // Release the HEAP node that held the packet just sent
    mBufPtr=getMQBufPtr(nextNetSend);
    mBufPtr=releaseHeap(mBufPtr);
    putMQBufPtr(nextNetSend,0);

    nextReply=nextNetSend;
    replyPending = true;

    netSendPending=false;
    nextNetSend=-1;

  } else if ( packetReceived ) {

    getMQ( lastNetRcvd, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for lastNetRcvd
    mID = mID | MQ_REPLY_MASK; // Set bit-13 to mark as REPLY
    mID = mID | MQ_LONG_MASK;  // Set bit-14 to mark as LONG
    mParErr=0;

    // Discover length of the received packet from the NET header stored in the first 8 bytes of HEAP[mBufPtr]

    mBufPtr=getMQBufPtr(lastNetRcvd);

    mLenPar2 = byte( heap[mBufPtr+ND_NBYT] );
    if ( byte( heap[mBufPtr+ND_TYPE] ) == NF_SBLK ) mLenPar2 = mLenPar2 * 4; // a large FSERVE block 4*ND_NBYT

    mLenPar2 += NET_HEADER_SZ; // don't forget the 8-byte packet header...

    netBytesRcvd += mLenPar2;

    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

    putMQ( lastNetRcvd, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

    nextReply=lastNetRcvd;
    netRecvPendCnt--;
    if ( netRecvPendCnt == 0 ) netRecvPending = false;
    replyPending = true;

  } // if...

  return; // all good - a Send or Recv was succesful and the respective message tagged as REPLY

} // actionNET()...

boolean waitGap(unsigned int gapTime) {
  // Wait for gapTime us, or return false if any net activity is detected

  timer=0; // Reset the us timer
  while ( timer < gapTime ) {

    if ( digitalRead(NET_IN) == HIGH ) return false; // Activity detected on the wire

  } // while...

  return true; // gapTime reached without any activity

} // waitGap...

boolean waitACTive(unsigned int timeOut, unsigned int minDuration) {
  // Wait for NET_IN going ACTive (or already ACTive) - cf. waitRisingEdge
  // - return true if ACTive found within timeOut us, of minDuration length
  // v2.1e - added check for minDuration=0, meaning we don't wait for ACTive level to persist

  unsigned int virtualTimer;
  boolean foundACTive=false;

  virtualTimer=0;

  while ( (virtualTimer < timeOut) && (foundACTive == false) ) {

    if ( PIND & 0b00000001 ) {

      foundACTive=true;
      PORTD |= 0b00000100; // DEBUG_NET2 HIGH

    } else {

      PORTD &= 0b11111011; // DEBUG_NET2 LOW

    } // if...

    virtualTimer++;

  } // while...

  if ( ! foundACTive  ) return false; // timeOut reached without detecting ACTive - error

  // v2.1e - leave early if we only wanted a level or an edge 
  if ( minDuration == 0 ) return true; // Detected ACTive within timeout - zero duration specified.

  virtualTimer=0;

  while ( (virtualTimer < minDuration) && (foundACTive == true) ) {

    if ( !(PIND & 0b00000001) ) {

      foundACTive=false;
      PORTD &= 0b11111011; // DEBUG_NET2 LOW

    } // if...

    virtualTimer++;

  } // while...

  if ( ! foundACTive ) return false; // timeOut reached without detecting ACTive - error

  return true; // Detected ACTive, within timeOut of at least minDuration

} // waitACTive...

boolean waitINACTive(unsigned int timeOut, unsigned int minDuration) {
  // Wait for NET_IN going INACTive (or already INACTive) - cf. waitFallingEdge
  // - return true if INACTive of minDuration length commenced within timeOut us
  // v2.1e - added check for minDuration=0, meaning we don't wait for INACTive level to persist

  unsigned int virtualTimer;
  boolean foundINACTive=false;

  virtualTimer=0;

  while ( (virtualTimer < timeOut) && (foundINACTive == false) ) {

    if ( PIND & 0b00000001 ) {

       // PORTD |= 0b00000010; // DEBUG_NET1 HIGH

    } else {

      foundINACTive=true;
      // PORTD &= 0b11111101; // DEBUG_NET1 LOW

    } // if...

    virtualTimer++;

  } // while...

  if ( ! foundINACTive  ) return false; // timeOut reached without detecting INACTive - error

  // v2.1e - leave early if we only wanted a level or an edge 
  if ( minDuration == 0 ) return true; // Detected INACTive within timeout - zero duration specified.

  virtualTimer=0;

  while ( (virtualTimer < minDuration) && (foundINACTive == true) ) {

    if ( (PIND & 0b00000001) ) {

      foundINACTive=false;
      // PORTD &= 0b11111101; // DEBUG_NET1 LOW

    } // if...

    virtualTimer++;

  } // while...

  if ( ! foundINACTive ) return false; // timeOut reached without detecting INACTive - error

  return true; // Detected INACTive, within timeOut of at least minDuration

} // waitINACTive...

boolean checkACTive(unsigned int minDuration) {
  // Check for NET_IN remaining ACTive for atleast minDuration

  unsigned int virtualTimer;
  boolean foundINACTive=false;

  // PORTD |= 0b00000010; // DEBUG_NET1 HIGH

  virtualTimer=0;

  while ( (virtualTimer < minDuration) && (foundINACTive == false) ) {

    if ( !(PIND & 0b00000001) ) {

      foundINACTive=true;
      PORTD &= 0b11111011; // DEBUG_NET2 LOW

    } else {

      PORTD |= 0b00000100; // DEBUG_NET2 HIGH

    } // if...

    virtualTimer++;

  } // while...

  // PORTD &= 0b11111101;

  if ( foundINACTive ) return false; // minDuration of ACTive not reached error

  return true; // Detected continuous ACTive of minDuration

} // checkACTive...

boolean waitRisingEdge(unsigned int timeOut) {
  // Wait for NET_IN going INACTive to ACTive (NOT already ACTive) - cf. waitACTive
  // - return true (ACTive edge found within timeOut us) or false

  unsigned int virtualTimer;
  boolean foundINACTive=false;
  boolean foundRisingEdge=false;

  virtualTimer=0; // Reset the us timer

  while ( virtualTimer < timeOut && (! foundINACTive) ) {

    if ( !(PIND & 0b00000001) ) foundINACTive=true;

    virtualTimer++;

  } // while...

  if ( ! foundINACTive ) return false; // INACTive not found in timeOut us

  while ( virtualTimer < timeOut && (foundRisingEdge == false) ) {

    if ( PIND & 0b00000001 ) foundRisingEdge=true; // Rising edge detected on the wire

    virtualTimer++;

  } // while...

  if ( foundRisingEdge ) {

    return true;

  } else {

    return false; // timeOut reached without Rising edge detected - error

  } // if...

} // waitRisingEdge...

boolean waitFallingEdge(unsigned int timeOut) {
  // Wait for NET_IN going ACTive to INACTive 
  // - return true (INACTive edge found within timeOut us) or false

  unsigned int virtualTimer;
  boolean foundACTive=false;
  boolean foundFallingEdge=false;

  virtualTimer=0;

  while ( (virtualTimer < timeOut) && (! foundACTive) ) {

    if ( PIND & 0b00000001 ) {

      foundACTive=true;
      PORTD |= 0b00000100; // DEBUG_NET2 HIGH

    } else {

      PORTD &= 0b11111011; // DEBUG_NET2 LOW

    } // if...

    virtualTimer++;

  } // while...

  if ( ! foundACTive ) return false; // ACTive not found in timeOut us

  while ( (virtualTimer < timeOut) && (foundFallingEdge == false) ) {

    if ( !(PIND & 0b00000001) ) {

      PORTD &= 0b11111011; // DEBUG_NET2 LOW
      foundFallingEdge=true; // Falling edge detected on the wire

    } else {

      // PORTD |= 0b00000010;

    } // if...

    virtualTimer++;

  } // while...

  if ( foundFallingEdge ) {

    return true;

  } else {

    return false; // timeOut reached without Falling edge detected - error

  } // if...

} // waitFallingEdge...

byte listenForPacket(unsigned int timeOut, int &recvMQSlot) {
  /* Listen for a packet on the wire for timeOut us
   * - depending upon the outcome, return the byte value:
   *   RECV_OK=0         - Packet received and chksums matched - ACK and Queue a REPLY
   *   RECV_TRY_AGAIN=1  - Some problem with RECV, but OK to retry - No ACK
   *   RECV_UNEXPECTED=2 - Packet HDR received OK, but did not match any pending receive - No ACK
   *   RECV_FAIL=4       - Packet received but either the header/data chksums didn't match, or some framing error
   *   RECV_TIMEOUT=5    - No packet detected in timeOut us
   * - For a successful receive, update the recvMQSlot (ref) parameter with the MQ Slot ID that matched the packet header.
  */

  boolean foundSCOUT, canSendACK, phantomBlk=false;

  byte rcvdType, rcvdNByt, rcvdDChk;
  byte recvCode, calcHChk, calcDChk=0;

  // v2.1b - Added FSERVE-style Data ChkSum vars
  unsigned int rcvdDChkFSERVE, calcDChkFSERVE=0;
  boolean flagFSERVE=false, flagBCast=false;

  byte hdrCheck;
  unsigned int packetLen;
  signed int mBufPtr;

  if ( netRecvPendCnt < 1 ) return RECV_UNEXPECTED; // Should NOT happen...

  noInterrupts();

  foundSCOUT=waitACTive(timeOut, SCOUT_DETECT_TIME);
  if ( ! foundSCOUT ) {

    PORTD |= 0b00010000; // DEBUG_NET4 - 15us: RECV_NO_SCOUT
    delayMicroseconds(15);
    PORTD &= 0b11101111;  

    interrupts();
    return RECV_TIMEOUT; // SCOUT not detected, or wasn't long enough

  } // if...

  delayMicroseconds(SCOUT_SKIP_TIME); // Skip-over the SCOUT (for now - I've got ideas...)

  recvCode = recvBytes(netHdrBuffer, NET_HEADER_SZ, calcHChk, calcDChkFSERVE); // NB calcHChk sums one too many bytes - correct in checkNETHeader.

  if ( recvCode != RECV_OK ) {

    PORTD |= 0b00010000; // DEBUG_NET4 - 25us: RECV_HDR_FRAME_ERROR
    delayMicroseconds(25);
    PORTD &= 0b11101111;

    interrupts();
    return RECV_TRY_AGAIN;

  } // if...

  // Capture and check Hdr ChkSum
  // v2.1b - Added FSERVE Data Chksum (.W) to checkNETHeader
  // hdrCheck = checkNETHeader( recvMQSlot, rcvdDst, rcvdSrc, rcvdBlkL, rcvdBlkH, rcvdType, rcvdNByt, rcvdDChk, rcvdHChk, calcHChk);
  
  hdrCheck = checkNETHeader( recvMQSlot, rcvdType, rcvdNByt, rcvdDChk, rcvdDChkFSERVE, flagFSERVE, flagBCast, calcHChk );

  switch (hdrCheck) {

    case HDR_GOOD:
      // The Packet Hdr was as anticipated and received without error (ChkSums matched)
      // Before ACKnowledging, check we have sufficient HEAP space to store the packet itself
      // - proceed to ACKnowledge (assuming NOT a BCAST) and read-in the actual packet payload.

      if ( rcvdType == NF_SBLK ) {

	packetLen = int(rcvdNByt) * 4; // This appears to be an FSERVE large Block - x4 (0 - 1020 bytes)

      } else {

	packetLen = int(rcvdNByt);

      } // if...

      if ( ((NET_HEADER_SZ+packetLen) > getMaxHeapSpc()) || (packetLen > NET_MAX_PACKET_SZ) ) {

	PORTD |= 0b00010000; // DEBUG_NET4 - 105us: RECV_OTHER_ERROR
	delayMicroseconds(105);
	PORTD &= 0b11101111;

	interrupts();
	return RECV_FAIL; // This isn't actually a recv error, but an unexpectedly long packet (>1020 bytes)...

      } // if...

      mBufPtr=allocateHeap(NET_HEADER_SZ+packetLen);
      putMQBufPtr(recvMQSlot, mBufPtr);

      // Copy over the received NET Header in to the start of new HEAP node
      copyArr(netHdrBuffer, &heap[mBufPtr], NET_HEADER_SZ);

      delayMicroseconds(PRE_ACK_DELAY);

      // Only send ACK byte for HDR if NOT receiving a Broadcast
      if ( ! flagBCast ) {

        sendByte(VALID_ACK, false); // Send ACK and leave INACTive

      } // if...
      break;

    case HDR_CHKSUM_ERROR:
     // A packet-hedaer arrived, but the ChkSum we calculated didn't match what we found in the Header
     // - we can't say for sure whether the packet was really destined for us or desired.
     //   Ignore and return.

      PORTD |= 0b00010000; // DEBUG_NET4 - 35us: RECV_HDR_CHKSUM_ERROR
      delayMicroseconds(35);
      PORTD &= 0b11101111;

      interrupts();
      return RECV_TRY_AGAIN;

    case HDR_NOT_FOR_ME:
      // The Packet Hdr was received but was either not for us (DST <> SELF)
      // - or else we weren't expecting a packet specifcally from SRC...
      //   Ignore and return.

      PORTD |= 0b00010000; // DEBUG_NET4 - 45us: RECV_HDR_NOT_FOR_ME
      delayMicroseconds(45);
      PORTD &= 0b11101111;

      interrupts();
      return RECV_UNEXPECTED;

    case HDR_PHANTOM_BLOCK:
      // A ghost from the past was received - a packet that we had previously wanted, received and ACKnowledged
      // - presumably, the sender failed to recognise our earlier ACK and has resent the (n-1) packet
      //   We'll ACK again patiently, but otherwise take no further action - quietly awaiting the next expected packet
      //  Note to self: we might want to leave the very last RECV message request for a given session 'open'
      //    for a time after ACK'ing, to accomodate this very situation and release the sender sooner than the standard QDOS protocol would...
      // v2.1e - Corrected Phantom Block handling - we need to read the packet and ACK it, before discarding it again

      PORTD |= 0b00010000; // DEBUG_NET4 - 55us: RECV_HDR_PHANTOM_BLOCK
      //PORTB |= 0b10000000; // DEBUG_NET6 - 55us: RECV_HDR_PHANTOM_BLOCK
      delayMicroseconds(55);
      PORTD &= 0b11101111;
      //PORTB &= 0b01111111;

      phantomBlk=true;
      packetLen = int(rcvdNByt);

      if ( ((NET_HEADER_SZ+packetLen) > getMaxHeapSpc()) || (packetLen > NET_MAX_PACKET_SZ) ) {

	PORTD |= 0b00010000; // DEBUG_NET4 - 105us: RECV_OTHER_ERROR
	delayMicroseconds(105);
	PORTD &= 0b11101111;

	interrupts();
	return RECV_FAIL; // This isn't actually a recv error, but that we don't have enough heap-space right now...

      } // if...

      // We need somewhere to receive this block, before we then discard it
      mBufPtr=allocateHeap(NET_HEADER_SZ+packetLen);

      // Only send ACK byte for HDR if NOT receiving a Broadcast
      if ( ! flagBCast ) {

        //delayMicroseconds(PRE_ACK_DELAY);
        sendByte(VALID_ACK, false); // Send ACK and leave INACTive

      } // if...
      break;

    case HDR_BLK_MISMATCH:
      // Any other BLK than a Phantom - this is a protocol-level error that should really trigger a RECV failure in QDOS

      //PORTD |= 0b00010000; // DEBUG_NET4 - 65us: RECV_HDR_BLK_MISMATCH
      PORTB |= 0b10000000; // DEBUG_NET6 -  65us: RECV_HDR_BLK_MISMATCH
      delayMicroseconds(65);
      //PORTD &= 0b11101111;
      PORTB &= 0b01111111;

      interrupts();
      return RECV_TIMEOUT;

    default:
      // Shouldn't get to this, but...   

      PORTD |= 0b00010000; // DEBUG_NET4 - 105us: RECV_OTHER_ERROR
      delayMicroseconds(105);
      PORTD &= 0b11101111;

      interrupts();
      return RECV_TIMEOUT;

  } // switch...

  // If we arrive here at all, we're ready to accept a packet meant for me (or a Broadcast).
  // - recvMQSlot will have been populated by checkNETHeader and point to the MQ appropriate for the REPLY
  // v2.1e - *** or this was a phantom block that we need to ACK (again!) before discarding

  // v2.1b - Modified recvBytes to accomodate FSERVE-type Data ChkSums
  recvCode = recvBytes(&heap[mBufPtr+NET_HEADER_SZ], packetLen, calcDChk, calcDChkFSERVE); // Read in to HEAP from NET_HEADER_SZ onwards

  if ( recvCode == RECV_FAIL ) {
    // Don't ACK (unless BCast) and simply return a read failure

    if ( flagBCast ) sendBCastNACK();

    mBufPtr=releaseHeap(mBufPtr);
    putMQBufPtr(recvMQSlot, 0);

    PORTD |= 0b00010000; // DEBUG_NET4 - 75us: RECV_DATA_FRAME_ERROR
    delayMicroseconds(75);
    PORTD &= 0b11101111;

    interrupts();
    return RECV_FAIL;

  } // if...

  // v2.1b - Modified Data ChkSum handling
  if ( ((flagFSERVE == true) && (calcDChkFSERVE != rcvdDChkFSERVE)) || ((flagFSERVE == false) && (calcDChk != rcvdDChk)) ) {
    // DATA ChkSum didn't match expected, return read failure (no ACK, unless BCast)...

    if ( flagBCast ) sendBCastNACK();

    mBufPtr=releaseHeap(mBufPtr);
    putMQBufPtr(recvMQSlot, 0);

    PORTD |= 0b00010000; // DEBUG_NET4 - 85us: RECV_DATA_CHKSUM_ERROR
    delayMicroseconds(85);
    PORTD &= 0b11101111;

    interrupts();
    return RECV_FAIL;

  } // if...

  // OK... So let's wait for end of the last STOP and then send the ACK, returning as read success.

  // Added v1.8h (following testing with ZXNet) - use waitINACTive in place of simple delay - before sending ACK
  // With the added check for a minimum gap
  // - setting reasonable defaults here - wait for a full bytes-worth of silence, starting within 3x BIT-times

  // canSendACK = waitINACTive( BIT_TIME_RECV * 10 , BIT_TIME_RECV * 10 );
  // v2.1b - Broadcast RECV is not likely to have this sort of INACTive period, (typically 7us) before another station might ACK
  //         so we'll just wait for the next Failling edge and then send our normal or BCast ACK anyway...

  canSendACK = waitINACTive( BIT_TIME_RECV * 10, BIT_TIME_RECV ); // Should always succeed, unless there is some other calamity...

  if ( ! canSendACK ) {

    mBufPtr=releaseHeap(mBufPtr);
    putMQBufPtr(recvMQSlot, 0);

    PORTD |= 0b00010000; // DEBUG_NET4 - 95us: RECV_DATA_ACK_NO_GAP
    delayMicroseconds(95);
    PORTD &= 0b11101111;

    interrupts();
    return RECV_FAIL;

  }

  // v2.1e - Added pre-ACK delay
  delayMicroseconds(PRE_ACK_DELAY);

  if ( flagBCast ) {

    sendBCastACK(); // Send Extended Broadcast ACK (5ms) and leave INACTive

  } else {

    sendByte(VALID_ACK, false); // Send ACK and leave INACTive

  } // if...

  // v2.1e - Discard the receive buffer if this was a Phantom block
  if ( phantomBlk ) {

    mBufPtr=releaseHeap(mBufPtr);
    interrupts(); // And the all-important re-enabling of interrupts...
    return RECV_TIMEOUT;

  } // if...

  // The REPLY itself (referenced by recvMQSlot) will be updated by the calling actionNet routine
  //   based on the packet header in HEAP at mBufPtr

  interrupts(); // And the all-important re-enabling of interrupts...
  return RECV_OK;

} // listenForPacket...

void sendBCastACK() {
  // v2.1b - Added Broadcast ACK and NACK handling

  digitalWrite(NET_OUT, NET_ACTive);

  delayMicroseconds(BCAST_ACK_TIME);

  digitalWrite(NET_OUT, NET_INACTive);

} // sendBCastACK

void sendBCastNACK() {
  // v2.1b - Added Broadcast ACK and NACK handling

  boolean gapFound=false, bAckFound=false;

  // Nominal timeout (4x bits-worth)
  gapFound = waitFallingEdge( BIT_TIME_RECV * 4 );

  // Wait 200us to check whether another station starts to send a BACK of their own
  bAckFound = waitACTive( BCAST_NACK_WAIT, BCAST_ACK_DETECT_TIME );
  
  if ( !bAckFound ) return;
  // No BACK from another receving station was seen in the alloted time
  //  so we are safe to presume that the Sending station will interpret that as a passive NACK anyway - no need to take further action

  // OK, so another station sent (what looks like) a Broadcast Ack, so we wait a little while after the line goes quiet before sending a counter NACK
  gapFound = waitFallingEdge( BCAST_ACK_EXT_TIME ); // we can discard this result, as we'll send our NACK anyway
  
  // Finally, send our own NACK to counter the other sender's ACK - after a short delay
  delay( BIT_TIME_RECV );
  sendBCastACK();

} // sendBCastNACK

byte checkNETHeader( int &matchingMQSlot, byte &typ, byte &nbt, byte &dck, unsigned int &dckFSERVE, boolean &flagFSERVE, boolean &bCastFlag, byte calcHChk) {
  // Capture the 8 bytes of netHdrBuffer in to (ref) params and check the Header ChkSum
  // - proceed to validate the DST, SRC, BLKL, BLKH against all the RECVieve-type MQ entries
  // v2.1b - Added 'Listen to self', 'Broadcast RECV' and FSERVE netHDR matching, as well as FSERVE blkL/blkH handling correction
  //         Modified Data ChkSum checking to accomodate FSERVE blocks and removed unneeded Ref params (moved to local vars)

  boolean foundRecvEntry=false;
  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum;
  
  byte dst, src, blkL, blkH, hck;

  dst  = netHdrBuffer[ND_DEST];
  src  = netHdrBuffer[ND_SELF];
  blkL = netHdrBuffer[ND_BLKL];
  blkH = netHdrBuffer[ND_BLKH];
  typ  = netHdrBuffer[ND_TYPE];
  nbt  = netHdrBuffer[ND_NBYT];
  dck  = netHdrBuffer[ND_DCHK];
  hck  = netHdrBuffer[ND_HCHK];

  calcHChk -= hck;

  if ( hck != calcHChk ) {

    return HDR_CHKSUM_ERROR;

  } // if...

  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {

    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for mQSlot

    if ( !( mDev == M_DEV_NET && ( ( mAct == M_ACT_NETRECV ) || ( mAct == M_ACT_NETSERV_START ) ) ) ) {

      // Not a RECV-type M_ID
      continue;

    } // if...

    // v2.1b - Added 'Listen to self' netHDR matching
    if ( (mAct == M_ACT_NETRECV) && (highByte(mParErr) == lowByte(mParErr)) ) {

      if (dst == highByte(mParErr)) {

        mParErr=256*word(dst) + word(src);
        putMQParErr(mQSlot, mParErr);
        // We are now ready to pass the next check...

      } // if...

    } // if...

    // v2.1b - Added 'Broadcast RECV' netHDR matching
    if ( (mAct == M_ACT_NETRECV) && (lowByte(mParErr) == 0) ) {

      mParErr=word(0, src);
      putMQParErr(mQSlot, mParErr);
      bCastFlag=true;
      // We are now ready to pass the next check...

    } // if...    

    if ( (mAct == M_ACT_NETRECV) && (word(dst, src) != mParErr) ) {

      // Didn't match this DST/SRC
      continue;

    } // if...
    
    // v2.1b - Added FSERVE netHDR matching
    if ( (mAct == M_ACT_NETSERV_START) && (dst != highByte(mParErr)) ) {
    
      // For an FSERVE type packet, we only check DST
      continue;

    } // if...

    foundRecvEntry=true;
    matchingMQSlot=mQSlot;
    break;

    // The structure of this for-loop is a bit messy due to the continue's effectively re-routing flow
    // - should probably be a do...while instead, but who's watching?

  } // for...

  if ( !foundRecvEntry ) {

    return HDR_NOT_FOR_ME;

  } // if...

  // Now, for ease of numeric comparison of BLK num (below is the ONLY such requirement),
  //   at both ends of the MQ (USB Host and here), we arrange for mLenPar2 to be in the big-endian form (BlkH, BlkL)
  //   rather than how they appear over the wire (which was originally to suit ZXNet running on the Z80)
  // Note to self PLEASE remember this when coding the MQ in SMSQ...

  // v2.1b - FSERVE blkL/blkH handling correction (FSERVE blocks don't use the Block index at all)
  if ( !(dst & 0b10000000) && !(src & 0b10000000) ) {
  
    if ( word(blkH, blkL) == (mLenPar2 - 1) ) {

      return HDR_PHANTOM_BLOCK;

    } // if...

    if ( word(blkH, blkL) != mLenPar2 ) {

      return HDR_BLK_MISMATCH;

    } // if...

  } // if...
  
  // v2.1b - Added check for FSERVE type blocks (bit7 set on either DST or SRC)
  if ( (dst & 0b10000000) || (src & 0b10000000) ) flagFSERVE = true;

  // Phew! Got there in the end. Ref param matchingMQSlot already updated and we can return read success

  return HDR_GOOD;

} // checkNETHeader...

byte recvBytes(byte *buffer, unsigned int byteCnt, byte &chkSum, unsigned int &chkSumFSERVE) {
  /* Listen for byteCnt bytes and populate buffer (HEAP or netHdrBuffer)
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP (+1x FILL between consecutive bytes)
   *   RETurns an error code - which we don't use much today, and two ref parmeters: chkSum & chkSumFSERVE
   *
   * v2.1b - Added Ref Param for FSERVE-style Data ChkSum
  */

  byte nextByte, recvCode;
  unsigned int leadWaitTime;

  leadWaitTime=LEAD_WAIT_TIME;

  chkSum = 0;

  for ( unsigned int bytePtr=0; bytePtr < byteCnt; bytePtr++ ) {

    recvCode = recvByte(leadWaitTime, nextByte);
    PORTD &= 0b11111011; // turn OFF DEBUG_NET2

    switch (recvCode) {

      case RECV_OK:

        buffer[bytePtr] = nextByte;
        chkSum += nextByte;
        
        // v2.1b - Added summation of FSERVE-style word-length Data ChkSum
        chkSumFSERVE +=nextByte;

        break;

      case RECV_FAIL:
        // Framing error (START or STOP not detected) - just fall-through to RECV_TIMEOUT...

      case RECV_TIMEOUT:
        // Timeout waiting for LEAD bits (could be less bytes in packet than there should be...)

        return recvCode; // Pass-back immediately to listenForPacket

    } // switch...

    leadWaitTime=LEAD_WAIT_TIME_2;

  } // for...

  return RECV_OK; // Wow! We successfully captured a whole block (HDR or DATA)!

} // recvBytes...

byte recvByte(unsigned int leadWaitTime, byte &receivedByte) {
  /* Wait for a ACTive LEAD bits in leadWaitTime (LEAD_WAIT_TIME for first byte/ACK, then nominal) us,
   * followed by INACTive START bit in START_WAIT_TIME
   * - 8x bits (LSBit first), followed by atleast 2 STOP bits (make cursory check, MIN checks for one, but Tk2 doesn't check at all)
   * - receivedByte returned by reference, error returned directly
  */

  boolean foundLEAD, foundSTART, foundSTOP;  
  // byte bitSampleCnt=0;
  // unsigned int virtualTimer;

  // v2.1e - adjust LEAD detection to just wait for ACTive, without waiting for longer than needed
  foundLEAD = waitACTive(leadWaitTime, 0); // Check just for ACTive
  if ( ! foundLEAD ) return RECV_TIMEOUT;

  foundSTART = waitFallingEdge(START_WAIT_TIME);
  if ( ! foundSTART ) return RECV_FAIL;

  delayMicroseconds(BIT0_TIME_RECV); // pause to centre of bit-0 (taking account of delay in detection of START bit)

  receivedByte=0; // reset the reference parameter

  // virtualTimer=0; // reset our virtual timer, ready to capture byte from NET_IN

  for ( byte b=1; b < 9; b++ ) {

    // for ( byte sample=0; sample < 5; sample++ ) {
    //  if ( digitalRead(NET_IN) == HIGH ) bitSampleCnt++;
    //}

    receivedByte = receivedByte >> 1; // shifted BEFORE adding next bit (avoids complications with 8x shifts)    
    PORTD |= 0b00000010; // turn ON DEBUG_NET1

    // if ( bitSampleCnt >= 3 ) receivedByte = receivedByte | 0b10000000; // NB Bit-7 set
    // if ( bitSampleCnt >= 1 ) {

    if ( PIND & 0b00000001 ) {

      receivedByte = receivedByte | 0b10000000; // NB Bit-7 set
      PORTD |= 0b00001000; // Switch ON DEBUG_NET3 (fast)

    } else {

      PORTD &= 0b11110111; // Switch OFF DEBUG_NET3 (fast)

    }

    // while ( timer < (b * BIT_TIME) ); // spin until next BIT_TIME expires

    // v1z9b: Extend time for Stobe DEBUG_NET1
    // delayMicroseconds(DEBUG_STROBE_TIME);
    PORTD &= 0b11111101; // turn OFF DEBUG_NET1
    // delayMicroseconds(BIT_TIME_RECV - DEBUG_STROBE_TIME);

    delayMicroseconds(BIT_TIME_RECV);

    //bitSampleCnt=0; // reset for next bit...

  } // for...

  PORTD &= 0b11110111; // turn OFF DEBUG_NET3 (finished sampling NetIN data bits)

  // We are now at the centre of first STOP bit

  // v2.1e - check only for a half-bit's worth of STOP
  foundSTOP = checkACTive(BIT_TIME_RECV / 2); // Check for at least 1/2 BIT_TIME of STOP
  if ( ! foundSTOP ) return RECV_FAIL;

  return RECV_OK; // Ref. parameter receivedByte has been created from bit samples

} // recvByte...

byte sendPacket(int sendMQSlot) {
  /* Attempt to transmit the packet referenced by sendMQSlot, with payload in HEAP
   * - depending upon the outcome, return the byte value:
   *   SEND_OK=0         - Packet sent and acknowledged by DST
   *   SEND_CONTENTION=1 - Contention detected during SCOUT phase
   *   SEND_NO_HACK=2    - Header sent, but was not acknowldged by DST
   *   SEND_NO_DACK=3    - Data packet sent, but was not acknowldged by DST
   * - GAP has already been found, so start with SCOUT (checking for contention) immediately...
   *
   * v2.1b - Broadcast SCOUT Extension handling
  */

  boolean scoutCode;
  byte mDev, mAct, recvCode, acknowledgeCode;
  unsigned int mID, mParErr, mLenPar2, mChkSum, packetLen;

  byte dst;
  boolean bCastFlag=false;
  
  signed int mBufPtr;

  getMQ( sendMQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for sendMQSlot

  mBufPtr=getMQBufPtr(sendMQSlot);
  dst=heap[mBufPtr + ND_DEST]; // Get the DST station number

  if (dst == 0) bCastFlag=true;

  scoutCode = sendSCOUT(SCOUT_WORD, bCastFlag);

  if ( ! scoutCode ) return SEND_CONTENTION; // Detected contention during SCOUT

  // Might want to extend SCOUT ACTive a little here - let's start without...

  noInterrupts();

  sendBytes(&heap[mBufPtr], NET_HEADER_SZ); // Transmit packet Header (8x bytes from index mBufPtr of HEAP)
  
  // v2.1b - Check for Broadcast send to determine whether or not we should wait for a Header ACK
  if ( bCastFlag == false ) {

    // Comment-out next if clause when testing - disables waiting for a HACK... 
    recvCode = recvByte(LEAD_WAIT_TIME, acknowledgeCode); // Presently only expecting 0x01 in response

    if ( recvCode != RECV_OK || acknowledgeCode != VALID_ACK ) {

      interrupts();
      return SEND_NO_HACK;

    } // if...

  } // if...

  // OK, got a positive/valid ACK (or this was a BCAST), let's pause a little before sending packet
  delayMicroseconds(POST_ACK_DELAY);

  packetLen = mLenPar2 - NET_HEADER_SZ;
  sendBytes(&heap[mBufPtr+NET_HEADER_SZ], packetLen); // Transmit packet Data (packetLen bytes from index mBufPtr+NET_HEADER_SZ of HEAP)

  // v2.1b - New routine to check for the Broadcast ACK/NACK called here (recv and acknowledgeCode are re-synthesised accordingly to match original design)
  if ( bCastFlag == true ) {

    recvCode = waitBroadcastACK(acknowledgeCode);

  } else {

    // Comment-out next if clause when testing - disables waiting for a DACK...
    recvCode = recvByte(LEAD_WAIT_TIME, acknowledgeCode); // Presently only expecting 0x01 in response

  } // if...

  interrupts();

  if ( recvCode != RECV_OK || acknowledgeCode != VALID_ACK ) { 

    return SEND_NO_DACK;

  } else {
    // Phew! Seems to have sent successfully - return OK

    netBytesSent += NET_HEADER_SZ + packetLen; // only update netBytesSent if packet was fully ACKnowledged.
    return SEND_OK;

  } // if...

} // sendPacket...

byte waitBroadcastACK(byte &ackCode) {
  /* v2.1b - new function to test for a TK2-style Broadcast ACK/NACK
  
     1. Wait for up to 500us for ACTive (ndt_back)
       2a. Timeout/ACTive not seen, so assume passive-NACK and restart
       2b. ACTive seen within 500us - continue to check for a subsequnet BNACK
     3. Wait for up to 20ms for INACTive (ndt_bace)
     4. After 20ms or going INACTive, whichever occurs first, wait 500us for next ACTive
       5a. Timeout/ACTive not seen, assume success
       5b. ACTive seen within 500us, interpret as active-NACK
  */
  
  boolean bAckFound=false, inActiveFound=false, bNAckFound=false;

  // 1. Wait for up to 500us for ACTive
  bAckFound=waitRisingEdge(BCAST_ACK_DETECT_TIME);

  if ( !bAckFound ) {

    // 2a. Timeout/ACTive not seen, so assume passive-NACK and restart
    ackCode=0;
    return RECV_FAIL;

  } // if...

  // 2b. ACTive seen within 500us - continue to check for a subsequnet BNACK

  // 3. Wait for up to 20ms for INACTive
  inActiveFound=waitFallingEdge(BCAST_ACK_EXT_TIME);

  // 4. After 20ms or INACTive, whichever occurs first, wait 500us for next ACTive
  bNAckFound=waitRisingEdge(BCAST_NACK_DETECT_TIME);

  if ( !bNAckFound ) {

    // 5a. Timeout/ACTive not seen, assume success
    ackCode = VALID_ACK;
    return RECV_OK;

  } else {

    // 5b. ACTive seen within 500us, interpret as active-NACK
    ackCode=0;
    return RECV_FAIL;

  } // if...

} // waitBroadcastACK

boolean sendSCOUT(unsigned int scoutWord, boolean bCastFlag) {
  /* Transmit SCOUT based on the scoutWord, checking for contention
   * - Uses SCOUT_BIT_TIME inplace of BIT_TIME
   *   scoutWord could be used in future implementations to flag capabilities for compatible clients
   *   Returns true if transmitted succesfully with no contention detected (only detectable durng NET_INACTive)
   *
   * v2.1b - Broadcast SCOUT Extension handling added
  */

  boolean scoutBitState;
  unsigned long scoutExtTime;

  scoutWord = scoutWord | 0b1000000000000011; // Ensure 2x START and 1x STOP

  timer=0; // Reset the us timer

  for ( byte scoutBit=0; scoutBit < SCOUT_BITS; scoutBit++ ) {

    scoutBitState = ! ( scoutWord & 0b0000000000000001 ); // NB Inverted logic suitable for PNP transitor    
    digitalWrite(NET_OUT, scoutBitState);

    if ( ! scoutBitState ) {

      // Can't check for contention during NET_ACTive
      delayMicroseconds(SCOUT_BIT_TIME);

    } else {

      // Check for contention whilst NET_INACTive, 5x times

      for ( byte chkContention=0; chkContention < 5; chkContention++ ) {

        delayMicroseconds(SCOUT_BIT_TIME_DIV5);      
        if ( digitalRead(NET_IN) == HIGH ) return false; // Contention detected, return false

      } // for...

    } // if...

    scoutWord = scoutWord >> 1;

  } // for...

  // Theoretically, 16x 30uS (SCOUT_BIT_TIME) = 480us have passed - we need to leave it ACTive until atleast 525us

  scoutExtTime = (SCOUT_DETECT_TIME + SCOUT_SKIP_TIME) - timer;

  // v2.1b - Broadcast SCOUT Extention handling added
  if (bCastFlag == true) scoutExtTime = scoutExtTime + SCOUT_BCAST_EXT_TIME;

  delayMicroseconds(scoutExtTime);

  return true; // SCOUT transmitted without contention being detected, return OK (NET_OUT still ACTive)

} // sendSCOUT...

void sendByte(byte nextByte, boolean leaveActive) {
  /* Transmit a single byte, leaving the NET_OUT port at the value of leaveActive
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP ( + 1x FILL if leaveActive )
   * - NB - NET_OUT needs the logical inverse using NET_ACTIve (false/LOW) and NET_INACTive (true/HIGH)
   *        due to the PNP transitor hardware used in the output circuit (matching QLAN design).
  */
  boolean bitState;

  timer=0;

  digitalWrite(NET_OUT, NET_ACTive);
  delayMicroseconds(LEAD_BITS * BIT_TIME_SEND);

  digitalWrite(NET_OUT, NET_INACTive);
  delayMicroseconds(START_BITS * BIT_TIME_SEND);

  for ( byte bit=0; bit < 8; bit++ ) {

    bitState = !(nextByte & 0b00000001); // NB Inverted logic suitable for PNP transitor
    digitalWrite(NET_OUT, bitState);
    delayMicroseconds(BIT_TIME_SEND);
    nextByte = nextByte >> 1;

  } // for...

  digitalWrite(NET_OUT, NET_ACTive);
  delayMicroseconds(STOP_BITS * BIT_TIME_SEND);

  if ( leaveActive ) {

    delayMicroseconds(FILL_BITS * BIT_TIME_SEND - 4);

  } else {

    digitalWrite(NET_OUT, NET_INACTive);

  } // if...

} //sendByte...

void sendBytes(byte *buffer, int byteCnt) {
  /* Transmit byteCnt bytes from buffer (HEAP)
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP (+1x FILL between consecutive bytes)
  */

  boolean leaveActive=true;
  byte nextByte;

  for ( int bytePtr=0; bytePtr < byteCnt; bytePtr++ ) {

    nextByte=buffer[bytePtr];
    if ( bytePtr == (byteCnt - 1) ) leaveActive=false;

    sendByte(nextByte, leaveActive);
    // Will add a FILL_BITS delay with NET_OUT left ACTive after each byte, except the very last

  } // for...

} // sendBytes..

void replyUSBMessage() {
  /* Send over USB the next REPLY from the MQ based on nextReply
  */

  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum;
  signed int mBufPtr;

  mQSlotAvailable=false;

  if ( nextReply == -1 ) {

    replyPending=false;
    return; // Make quick return if, for some reason, the nextReply is not set (-1)

  } // if...

  // nextReply should therefore hold the mQSlotIndex in to the Message Queue, ready for REPLY

  getMQ( nextReply, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // Get some details about the REPLY

  // Serial.flush(); // On Teensy, will wait until all previous but still pending Sends are completed, or return immediately

  flashLED_start(LED_PIN_USB_TX);

  Serial.write( MQ_SYNC, MQ_SYNC_LEN );
  Serial.write( &mQTable[nextReply * MQTABLE_ENTRY_SZ], MESSAGE_HLEN );

  if ( (mID & MQ_LONG_MASK) && (mLenPar2 > 0) ) {

    mBufPtr=getMQBufPtr(nextReply);

    Serial.write( &heap[mBufPtr], mLenPar2 );
//    usbSendBuffersFree++;
//    usbSendBufferFree=true;

    mBufPtr=releaseHeap(mBufPtr); // Release the heap-space used for the LONG reply

  }

  Serial.send_now(); // On Teensy, will force (and wait for) all pending USB data to be transmitted

  usbBytesSent+= MQ_SYNC_LEN + MESSAGE_HLEN + mLenPar2;

  markMQSlotEmpty( nextReply ); // Flag this MQ Slot entry to show processed/now empty

  // Find next free MQ Slot (dupe of what we do in checkUSBMessage)
  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {

    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );

    if ( mID == 0 ) {

      nextFreeMQSlot=mQSlot; // Find the lowest free MQ Slot
      mQSlotAvailable=true;
      break; // This slot is 'empty' - we're done, leave for loop

    } // if...

  } // for...

  nextReply = -1;
  replyPending = false;

} // replyUSBMessage()...

void copyArr( byte *srcArr, byte *dstArr, int byteCnt ) {

  for ( int byteIdx=0; byteIdx < byteCnt; byteIdx++ ) dstArr[byteIdx] = srcArr[byteIdx];

} // copyArr()...

void copyArray( byte *srcArr, byte *dstArr, int srcIdx, int dstIdx, int byteCnt ) {

  int byteIdx;

  for ( byteIdx=0; byteIdx < byteCnt; byteIdx++ ) {
    dstArr[dstIdx] = srcArr[srcIdx];
    dstIdx++;
    srcIdx++;
  }

} // copyArray()...

void loop() {
  // put your main code here, to run repeatedly:

  initialiseVariables();

  while (!Serial.dtr()) ; // Wait for program running on host PC to open COM port

  flashLEDs(3, 250);

  do {

    checkUSBMessage();

    if ( mQScanRequired ) {

      scanMessageQueue();
// flashLEDs(1, 250);
// flashLEDs(1, 500);

    } // if...

    if ( netActionPending ) {

      actionNET();
// flashLEDs(3, 250);
// flashLEDs(1, 500);            
    } // if...

    if ( replyPending ) {

      replyUSBMessage();
// flashLEDs(5, 250);
// flashLEDs(1, 500);

    } // if...

    flashLED_async();

  } while (cmd != M_ACT_BRIDGE_RESET);

} // loop()...
